#!/usr/bin/env python

#
# CA Explorer program for exploring cellular automata
# Copyright (C) 2009 Lauritz Thaulow
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

START_MESSAGE = """
  CA Explorer is Copyright (C) 2009 Lauritz Thaulow
  This program comes with ABSOLUTELY NO WARRANTY; for details,
  read the "About" dialog.

  This is free software, and you are welcome to redistribute it
  under certain conditions; see the "About" dialog for the license.
"""

VERSION = "0.2"
FRAME_SIZE = (800, 600)

import wx
import os
import sys
from math import log10, ceil
from time import time
from types import TupleType, IntType, FloatType
from fractions import Fraction
from cca.cca import *
import numpy as np

ID_SAVE = 102
ID_ABOUT = 101
ID_EXIT = 110

DEFAULT_MAX_ITEMS = 20

CA_CHOICES = [
    ("1D 1R binary", CA1D_bin, "LatticeCache"),
    ("1D totalistic", CA1D_tot, "LatticeCache"),
    ("Kusch-Markus", KuschMarkus, "LatticeCache"),
    ]

DIV_HEIGHT = 100



class BinDict(dict):
    """A prototype class for a list with "bins".

This class has the set_div method which sets how
the keys are divided into bins. By default, div is
set to 10:
>>> bd = BinDict()
>>> bd[18] = 3
>>> bd[19] = 5
>>> bd[25] = 8
>>> bd[30] = 10
>>> bd
{10: 5, 20: 8, 30: 10}
>>> bd.set_div(3)
>>> bd
{}
>>> for i in range(15):
...   bd[i] = i
... 
>>> bd
{0: 2, 9: 11, 3: 5, 12: 14, 6: 8}
>>> bd.get_key(5)
3
>>> bd.has_key(5)
True
>>> bd.has_key(30)
False
>>> 
"""
    def __init__(self, div=10):
        self.set_div(div)

    def reset(self):
        self.clear()

    def set_div(self, div):
        self.div = int(div)
        self.reset()

    def has_key(self, i):
        return dict.has_key(self, self.get_key(i))

    def get_key(self, i):
        return int(i) - int(i) % self.div

    def next_key(self, i):
        return self.get_key(i) + self.div

    def bin_start(self, i):
        return self.div*self.get_key(i)

    def __getitem__(self, i):
        return dict.__getitem__(self, self.get_key(i))

    def __setitem__(self, i, value):
        return dict.__setitem__(self, self.get_key(i), value)


class StateStore(BinDict):
    """Stores the state of a 1D CA at regular intervals.

Works like a dict, the value of which is an ndarray of type
"uint8" with shape (h+1,w,3), where (w,h) can be set with the
set_shape method.
"""
    def __init__(self, ca, max_items):
        self.set_ca(ca)
        BinDict.__init__(self, ca.steps-1)

    def reset(self):
        BinDict.reset(self)
        self.ca.init_arrs(reset=True)

    def set_div(self, n):
        BinDict.set_div(self, n)

    def set_ca(self, ca):
        self.ca = ca
        self.set_div(ca.steps-1)

    def __getitem__(self, row):
        if len(self) == 0:
            self.ca.init_arrs()
            self[0] = self.ca.state
        val = BinDict.__getitem__(self, row)
        self.ca.state = val
        self.ca.run()
        nextrow = row + self.div
        if not self.has_key(nextrow):
            self[nextrow] = self.ca.state
        #sys.stdout.write("Store(%d), " % row)
        #sys.stdout.flush()
        return (self.get_key(row), self.ca.get_img())


class Cache(BinDict):
    """A class for implementing a tiered cache system.

This is a prototype class, subclasses need to define the 
class variables MAX_ITEMS and underlay_class, and the
methods process_key and process.

MAX_ITEMS is a list of the number of items to be stored
in the cache down the chain. The last element of the list
will be popped off and used, the rest of the list sent to
the tiers below. If the list is too short the last remaining
element will be used by all the remaining tiers.

underlay_class is the class that should be instanced to 
act as the next tier in the cache pyramid. It should be a
subclass of BinDict with the methods reset and set_size.

process_key should return a valid key for the underlay,
given a valid key for the current class.

process should prepare and return a given element of the 
underlay class for storage in the current class.
"""
    MAX_ITEMS = [100]
    underlay_class = None

    def __init__(self, ca, max_items=None):
        if self.__class__.underlay_class == None:
            raise RuntimeError("You must set the class variable " +
                               "underlay_class to something sensible.")
        (this_max, next_max) = self.process_max(max_items)
        self.max_items = this_max
        self.underlay = self.__class__.underlay_class(ca, next_max)
        BinDict.__init__(self, self.underlay.div)
        self.ca = ca

    def process_max(self, max_items):
        if max_items == None:
            max_items = self.__class__.MAX_ITEMS
        if len(max_items) > 1:
            this = max_items.pop()
            return (this, max_items)
        return (max_items[0], max_items)

    def reset(self):
        BinDict.reset(self)
        self.underlay.reset()
        self.timestamp = {}

    def set_ca(self, ca):
        self.ca = ca
        self.underlay.set_ca(ca)

    def process(self, key):
        """Fetch and prepare an item from the underlay for storage.

The argument key should be used to get an item from the underlay,
enhanche or specialize it in some way, and then return that.
This method is called by __getitem__ when it needs to store an
item in this cache.
"""
        return self.underlay[self.process_key(key)]

    def process_key(self, key):
        """Translate the key for this tier to a key for the underlay.

Given a valid key for the current class, it should return a valid
key for the underlay class.
"""
        return key

    def __getitem__(self, key):
        if self.ca.has_changed():
            self.reset()
        if not self.has_key(key):
            val = self.process(key)
            self[key] = val
        else:
            #sys.stdout.write("%s(%s), " % \
            #    (self.__class__.__name__, repr(key)))
            #sys.stdout.flush()
            val = BinDict.__getitem__(self, key)
        self.timestamp[self.get_key(key)] = time()
        return val

    def __setitem__(self, key, val):
        BinDict.__setitem__(self, key, val)
        self.timestamp[self.get_key(key)] = time()
        if len(self) > self.max_items:
            oldkey = sorted(self.keys(), 
                            key=self.timestamp.__getitem__)[0]
            del self[oldkey]
            del self.timestamp[oldkey]


class Rect:
    """Integer rectangle with operations."""
    def __init__(self, x, y, w, h):
        if w <= 0 or h <= 0:
            raise RuntimeError("Invalid width %d or height %d." % (w, h))
        (self.x, self.y) = (x, y)
        (self.w, self.h) = (w, h)
        self.coerce()

    def coerce(self):
        (self.x, self.y) = (int(self.x), int(self.y))
        (self.w, self.h) = (int(self.w), int(self.h))
        
    def __and__(self, rect):
        if rect == None:
            return None
        (a, b) = (rect, self)
        if a.x < b.x:
            (a, b) = (b, a)
        if b.x + b.w <= a.x:
            return None
        if a.y < b.y:
            (a, b) = (b, a)
        if b.y + b.h <= a.y:
            return None
        # Now we know they overlap and a.y > b.y
        y = a.y
        h = min(a.h, b.y + b.h - a.y)
        if a.x < b.x:
            (a, b) = (b, a)
        x = a.x
        w = min(a.w, b.x + b.w - a.x)
        #print "%s & %s = %s" % (self, rect, "(%d, %d, %d, %d)" % (x,y,w,h))
        ret = Rect(x, y, w, h)
        return ret
    __rand__ = __and__

    def __eq__(self, r):
        if r == None:
            return False
        if r.x == self.x and r.y == self.y and \
                r.w == self.w and r.h == self.h:
            return True
        return False
    __req__ = __eq__

    def copy(self):
        return Rect(self.x, self.y, self.w, self.h)

    def wxRect(self):
        return wx.Rect(self.x, self.y, self.w, self.h)

    def shatter(self, i):
        o = self.copy()  # The "outer" rectangle, the argument is the "inner"
        n = o & i   
        shift = 0
        if n == None:
            o.x = o.x % i.w
            shift = self.x - o.x
            n = o & i
        if n == None:
            return (None, [o])
        i = n  # We only care about the part of i inside o
        pieces = []
        if o.x < i.x:
            w = i.x - o.x
            pieces.append(Rect(o.x + shift, o.y, w, o.h))
            o.x += w; o.w -= w
        # Now we know the left sides of i and o align
        if o.y < i.y:
            h = i.y - o.y
            pieces.append(Rect(o.x + shift, o.y, o.w, h))
            o.y += h; o.h -= h
        # Now i is in the upper left corner of o
        if o.w > i.w:
            pieces.append(Rect(i.x + i.w + shift, o.y, o.w - i.w, o.h))
            o.w = i.w
        # Now i fills the whole upper part of o
        if o.h > i.h:
            pieces.append(Rect(o.x + shift, i.y + i.h, o.w, o.h - i.h))
        i.x += shift
        #print "Shattered %s into %s with %s" % (self, pieces, i)
        return (i, pieces)

    def translate(self, fromref, toref):
        x = (self.x - fromref.x)*toref.w/fromref.w + toref.x
        y = (self.y - fromref.y)*toref.h/fromref.h + toref.y
        w = self.w*toref.w/fromref.w
        h = self.h*toref.h/fromref.h
        return Rect(int(x), int(y), int(w), int(h))

    def __repr__(self):
        return "Rect(%d, %d, %d, %d)" % (self.x, self.y, self.w, self.h)
    __str__ = __repr__



class ImageChunk: 
    def __init__(self, img, ul):
        (x, y) = ul
        (h, w, dummy) = img.shape
        h -= 1
        #self.im = wx.NullImage()
        #self.im.Create(w, h, clear=False)
        self.im = wx.EmptyImage(w, h)
        view = img[0:h,...]
        self.im.SetData(view.tostring())
        self.rect = Rect(x, y, w, h)

    def GetSubImage(self, rect):
        if rect == self.rect:
            #print "Returning copy"
            return self.im.Copy()
        r = rect.copy()
        r.x = r.x % self.rect.w
        n = self.rect & r
        if n == None:
            return None
        n.y -= self.rect.y
        return self.im.GetSubImage(n.wxRect())

    def chip(self, rect):
        return rect.shatter(self.rect)

    def get_shape(self):
        return (self.rect.w, self.rect.h)
    shape = property(get_shape)


class LatticeCache(Cache):
    MAX_ITEMS = [50]
    underlay_class = StateStore

    def process(self, row):
        (key, img) = self.underlay[row]
        #print img.shape
        return ImageChunk(img, (0, key))


class ZoomCache(Cache):
    MAX_ITEMS = [50, 20]
    underlay_class = LatticeCache

    def process_key(self, key):
        return key[0]

    def get_key(self, cache_key):
        (key, zoom) = cache_key
        return (Cache.get_key(key), zoom)

    def next_key(self, cache_key):
        (key, zoom) = cache_key
        return (Cache.next_key(key), zoom)

    def process(self, cache_key):
        "Creates the item to be stored from the underlay."
        (key, im) = self.underlay[self.process_key(cache_key)]
        zoom = cache_key[1]
        if zoom == 1.0:
            return (key, zoom, im)
        (w, h) = (im.GetWidth(), im.GetHeight())
        (nw, nh) = (int(w*zoom), int(h*zoom))
        if zoom < 1:
            w -= w % nw
            h -= h % nh
            im.Resize(w, h)
        im = im.Scale(nw, nh)
        return (index_key, zoom, im.ConvertToBitmap())


class Animation(StateStore): 
    def __init__(self, ca, max_items=None):
        StateStore.__init__(self, ca, max_items)
        self.current_img = None
        self.current_step = 0
        
    def __getitem__(self, step):
        if self.current_img == None:
            self.current_img = ImageChunk(self.ca.get_img(step), (0,0))
        if step <= self.current_step:
            return self.current_img
        self.ca.run()
        self.current_img = ImageChunk(self.ca.get_img(step), (0,0))            
        self.current_step += 1
        return self.current_img


class CA_Drawing:
    def __init__(self, chunk_size=200):
        self.chunk_size = chunk_size
        self.cache = None
        self.ca = None

    def set_ca(self, ca, interface):
        self.ca = ca
        if issubclass(interface, Animation):
            self.chunk_size = 1
        else:
            self.chunk_size = 200
        ca.steps = self.chunk_size + 1
        self.cache = interface(ca)

    def Blit(self, dc, src, rects, zoom, step=None):
        bm = None
        memdc = None
        while len(rects) != 0:
            rect = rects.pop()
            if step != None:
                im = self.cache[step]
            else:
                im = self.cache[rect.y]
            (r, whatsleft) = im.chip(rect)
            if r == None:
                continue
            rects.extend(whatsleft)
            if zoom != 1:
                if zoom < 1:
                    try:
                        dw = r.w % int(r.w*zoom)
                        dh = r.h % int(r.h*zoom)
                    except ZeroDivisionError:
                        continue
                    if dw != 0 or dh != 0:
                        r.w -= dw
                        r.h -= dh
            piece = im.GetSubImage(r)
            if zoom != 1:
                piece.Rescale(int(r.w*zoom), int(r.h*zoom))
            if memdc != None:
                del memdc
            memdc = wx.MemoryDC(piece.ConvertToBitmap())
            tx = int((r.x - src.x)*zoom)
            ty = int((r.y - src.y)*zoom)
            (w, h) = (int(r.w*zoom), int(r.h*zoom))
            dc.Blit(tx, ty, w, h, memdc, 0, 0)
            #dc.SetPen(wx.WHITE_PEN)
            #dc.SetTextForeground(wx.WHITE)
            #dc.SetTextBackground(wx.BLACK)
            #dc.SetBrush(wx.TRANSPARENT_BRUSH)
            #dc.DrawRectangle(tx, ty, w, h)
            #dc.DrawText("%d %d %d %d" % (tx, ty, w, h), tx+3, ty+3)

    def reset(self):
        self.set_width(self.width, True)



class ParamTextCtrl(wx.TextCtrl):
    def __init__(self, parent, name, ca):
        self.ca = ca
        self.pp = parent.pp
        self.name = name
        self.controller = ca[name]
        self.history = []
        #self.controller.on_change.append(self.force_redraw)
        wx.TextCtrl.__init__(self, parent, wx.ID_ANY, 
                             self.controller.getstring(self.ca),
                             (0, 0), (70, -1), 
                             wx.TE_RIGHT|wx.TAB_TRAVERSAL)
        self.Bind(wx.EVT_KILL_FOCUS, self.update)
        #self.Bind(wx.EVT_TEXT_ENTER, self.CheckChanged)
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnWheel)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def update(self, e=None):
        setattr(self.ca, self.name, self.GetValue())
        if self.controller.has_changed(self.ca, reset=False):
            self.SetValue(self.controller.getstring(self.ca))
        
    def OnChar(self, e):
        kc = e.GetKeyCode()
        if kc in (9, 13):
            self.update()        
        elif kc in (82, 114):  # Rr
            if hasattr(self.controller, "randomize"):
                self.history.append(getattr(self.ca, self.name))
                self.controller.randomize(self.ca)
                self.SetValue(self.controller.getstring(self.ca))
        elif kc == 26:  # ctrl-z
            if self.history:
                val = repr(self.history.pop())
                self.SetValue(val)
                print val
                self.update()
        else:
            e.Skip()
            return
        self.force_redraw()

    def OnWheel(self, e):
        rot = e.GetWheelRotation()
        if rot == 0:
            return
        steps = MEDIUM_STEP
        if e.ShiftDown():
            steps = SMALL_STEP
        elif e.ControlDown():
            steps = LARGE_STEP
        self.controller.step(self.ca, rot, steps)
        self.SetValue(self.controller.getstring(self.ca))
        self.force_redraw()

    def force_redraw(self, vars=None):
        self.pp.Reset()
        self.pp.PaintPattern()


class ParamCtrl(wx.BoxSizer):
    def __init__(self, parent, name, ca):
        wx.BoxSizer.__init__(self, wx.VERTICAL)
        self.ctrl = ParamTextCtrl(parent, name, ca)
        self.Add(wx.StaticText(parent, wx.ID_ANY, name))
        self.Add(self.ctrl)        

    def MoveAfterInTabOrder(self, other):
        self.ctrl.MoveAfterInTabOrder(other.ctrl)


class PatternPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, wx.ID_ANY, size=FRAME_SIZE, style=wx.BG_STYLE_CUSTOM|wx.TAB_TRAVERSAL)
        self.SetBackgroundColour('white')
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnWheel)
        self.Bind(wx.EVT_MOTION, self.OnMouse)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.ca_draw = CA_Drawing()
        self.y = 0
        self.x = 0
        self.step = 0
        self.zoom = Fraction(1,1)
        self.state = None
        self.delay = None

    def set_ca(self, ca, interface):
        self.ca_draw.set_ca(ca, interface)
        self.OnResize(None)

    def Prepare(self):
        (w, h) = (self.buf.GetWidth(), self.buf.GetHeight())
        screen = Rect(0, 0, w, h)
        (nx, ny, nz) = (self.x, self.y, self.zoom)
        (nw, nh) = (int(w/nz), int(h/nz))
        newview = Rect(nx, ny, nw, nh)
        if self.state != None and nz/self.state[2] == 1:
            dc = wx.BufferedDC(None, self.buf)
            (ox, oy, oz, st) = self.state
            factor = nz/oz
            (ow, oh) = (int(w/oz), int(h/oz))
            oldview = Rect(ox, oy, ow, oh)
            common = oldview & newview
            if common == None or st != self.step:
                return (dc, newview, [newview])
            frombox = common.translate(oldview, screen)
            tobox = common.translate(newview, screen)
            im = self.buf.ConvertToImage()
            im = im.GetSubImage(frombox.wxRect())
            if factor != 1:
                im.Rescale(tobox.w, tobox.h)
            bm = im.ConvertToBitmap()
            memdc = wx.MemoryDC(bm)
            dc.Blit(tobox.x, tobox.y, tobox.w, tobox.h, memdc, 0, 0)
            #dc.SetPen(wx.WHITE_PEN)
            #dc.SetBrush(wx.TRANSPARENT_BRUSH)
            #dc.DrawRectangle(target.x, target.y, target.w, target.h)
            (dummy, todraw) = newview.shatter(common)
            return (dc, newview, todraw)
        return (wx.BufferedDC(None, self.buf), newview, [newview])


    def PaintPattern(self):
        (dc, view, todraw) = self.Prepare()
        if issubclass(self.ca_draw.cache.__class__, Animation):
            self.ca_draw.Blit(dc, view, todraw, self.zoom, self.step)
        else:
            self.ca_draw.Blit(dc, view, todraw, self.zoom)
        del dc
        self.OnPaint(None)
        self.state = (self.x, self.y, self.zoom, self.step)

    def OnResize(self, evt):
        width, height = self.GetClientSizeTuple()        
        self.buf = wx.EmptyBitmap(width, height)
        self.Reset()

    def Reset(self):
        width, height = self.GetClientSizeTuple()
        self.y = 0
        self.x = self.ca_draw.ca.width/2 - width/2
        self.step = 0
        self.zoom = Fraction(1/1)
        self.state = None
        self.PaintPattern()

    def OnPaint(self, evt):
        if not hasattr(self, "buf"):
            self.OnResize(None)
        wx.BufferedPaintDC(self, self.buf)

    def OnMouse(self, evt):
        (x,y) = evt.GetPosition()
        xm = (self.x + x/self.zoom) % self.ca_draw.ca.width
        ym = self.y + y/self.zoom
        parent = self.GetParent()
        parent.SetStatusText("%d, %d" % (xm, ym))
        if evt.RightIsDown():
            self.ca_draw.ca.arr[...,int(ym),int(xm)] = 1
            
        
    def OnWheel(self, evt):
        width, height = self.GetClientSizeTuple()
        delta = -evt.GetWheelRotation()/(self.zoom*2) or 1
        if evt.ShiftDown():
            self.x = (self.x + delta) % self.ca_draw.ca.width
        elif evt.ControlDown():
            (width, height) = self.GetClientSizeTuple()
            (x,y) = evt.GetPosition()
            (xm,ym) = (x/self.zoom, y/self.zoom)
            if (delta > 0) and (self.zoom > 0.12):
                self.x = (self.x - xm) % self.ca_draw.ca.width
                self.y = self.y - ym
                self.zoom /= 2
            elif (delta < 0) and (self.zoom < 8):
                self.x = (self.x + xm/2) % self.ca_draw.ca.width
                self.y = self.y + ym/2
                self.zoom *= 2
        else:
            self.y += delta
        if self.y < 0:
            self.y = 0
        self.PaintPattern()

    def OnChar(self, e):
        kc = e.GetKeyCode()
        if kc == 32:
            self.step += 1
            self.delay = None
        elif kc in (82, 114):  # Rr
            self.ca_draw.ca.reset()
        self.PaintPattern()
        

class AboutDialog(wx.Dialog):
    def __init__(self):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "About CA Explorer", 
                           size=FRAME_SIZE, style=wx.DEFAULT_DIALOG_STYLE)
        sizer = wx.BoxSizer(wx.VERTICAL)
        text = "\nCA Explorer %s is (C) 2009 Lauritz Thaulow\n" % VERSION
        sizer.Add(wx.StaticText(self, wx.ID_ANY, text))
        path = os.path.dirname(sys.argv[0])
        license = open(os.path.join(path, "LICENSE.txt"), "r").read()
        content = wx.TextCtrl(self, wx.ID_ANY, license, (0,0), (640, -1),
                              style=wx.TE_READONLY|wx.TE_MULTILINE)
        content.SetFont(wx.Font(10, wx.FONTFAMILY_TELETYPE, 
                                wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL))
        sizer.Add(content, 1, wx.EXPAND)
        ok = wx.Button(self, wx.ID_ANY, "OK")
        self.Bind(wx.EVT_BUTTON, self.OnOK)
        sizer.Add(ok, 0, wx.ALIGN_CENTER)
        self.SetSizer(sizer)
        content.SetInsertionPoint(0)
        
    def OnOK(self, evt):
        self.Destroy()

class RectDialog(wx.Dialog):
    def __init__(self, vars):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "Specify dimensions", 
                           size=(640, 480), 
                           style=wx.DEFAULT_DIALOG_STYLE)
        self.vars = vars
        self.inp = {}
        sizer = wx.BoxSizer(wx.VERTICAL)
        for v in vars:
            sizer2 = wx.BoxSizer(wx.HORIZONTAL)            
            sizer2.Add(wx.StaticText(self, wx.ID_ANY, "%s: " % v))
            self.inp[v] = wx.TextCtrl(self, wx.ID_ANY, "%d" % self.vars[v])
            sizer2.Add(self.inp[v])
            sizer.Add(sizer2)
        ok = wx.Button(self, wx.ID_ANY, "OK")
        self.Bind(wx.EVT_BUTTON, self.OnOK)
        sizer.Add(ok, 0, wx.ALIGN_CENTER)
        self.SetSizer(sizer)
        
    def OnOK(self, evt):
        for v in self.vars:
            self.vars[v] = int(self.inp[v].GetValue())
        self.Destroy()

class MainWindow(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, size=(800, 600))
        self.args = {}
        self.changed = False
        self.curr_clss = None
        filemenu= wx.Menu()
        filemenu.Append(ID_SAVE, "&Save..."," Save parts of this image")
        filemenu.Append(ID_ABOUT, "&About"," Information about this program")
        filemenu.AppendSeparator()
        filemenu.Append(ID_EXIT,"&Quit"," Exit the program.")
        menuBar = wx.MenuBar()
        menuBar.Append(filemenu,"&File")
        self.SetMenuBar(menuBar)

        self.pp = PatternPanel(self)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        self.ca_choice = wx.Choice(self, wx.ID_ANY, wx.Point(0,0),
                                   wx.Size(150,30), 
                                   [x[0] for x in CA_CHOICES])
        sizer2.Add(self.ca_choice)
        self.param_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer2.Add(self.param_sizer)
        self.OnChoice(None)
        self.sizer.Add(sizer2)
        self.sizer.Add(self.pp, 1, wx.EXPAND)
        self.SetAutoLayout(True)
        self.SetSizer(self.sizer)
        self.Layout()
        #self.sizer.Fit(self)

        self.SetMinSize(wx.Size(10,10))
        self.CreateStatusBar()
                  
        wx.EVT_MENU(self, ID_SAVE, self.OnSave)
        wx.EVT_MENU(self, ID_ABOUT, self.OnAbout)
        wx.EVT_MENU(self, ID_EXIT, self.OnExit)
        #self.Bind(wx.EVT_IDLE, self.OnIdle)

        self.ca_choice.Bind(wx.EVT_CHOICE, self.OnChoice)
        self.Show(True)
        self.pp.Bind(wx.EVT_SIZE, self.pp.OnResize)
        self.SetFocus()
        #self.pp.OnResize(None)
        #self.pp.PaintPattern(self.args)

    def OnSave(self,e):
        file = wx.FileDialog(self, "Specify filename to save to", 
                             "", "ca.png", "*.png", 
                             wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        status = file.ShowModal()
        if status != wx.ID_OK:
            return
        filename = file.GetPath()
        file.Destroy()
        maxw = self.pp.ca_draw.ca.width
        vars = {"x":0, "y":0, "w":maxw, "h":0}
        getrect = RectDialog(vars)
        getrect.ShowModal()
        if vars["w"] > maxw:
            vars["w"] = maxw
        buf = wx.EmptyBitmap(vars["w"], vars["h"])
        memdc = wx.MemoryDC(buf)
        view = Rect(vars["x"], vars["y"], vars["w"], vars["h"])
        self.pp.ca_draw.Blit(memdc, view, [view], 1)
        del memdc
        buf.SaveFile(filename, wx.BITMAP_TYPE_PNG)
    

    def OnChoice(self, e):
        (name, ca_class, interface) = \
            CA_CHOICES[self.ca_choice.GetCurrentSelection()]
        if self.curr_clss == ca_class:
            return
        self.curr_clss = ca_class
        ca = ca_class()
        self.param_sizer.DeleteWindows()
        ctrls = [ParamCtrl(self, i, ca) for i in ca \
                     if not ca[i].hidden]
        for i in range(len(ctrls)):
            self.param_sizer.Add(ctrls[i])
            ctrls[i].MoveAfterInTabOrder(ctrls[(i-1) % len(ctrls)])
        self.Layout()
        #self.sizer.Fit(self)
        self.pp.set_ca(ca, eval(interface))
        #print self.GetChildren()

    def OnAbout(self,e):
        d = AboutDialog()
        d.ShowModal()
        d.Destroy()

    def OnExit(self,e):
        self.Close(True)  # Close the frame.


def start():
    print START_MESSAGE
    import sys,os
    app = wx.PySimpleApp()
    frame = MainWindow(None, wx.ID_ANY, "CA Explorer")
    app.MainLoop()


if __name__ == '__main__':
    #import cProfile
    #cProfile.run('start()', 'profile.txt')
    start()


