//
// _ccamodule.c, A python extension module for fast CA calculation
// Copyright (C) 2009 Lauritz Thaulow
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <Python.h>
#define PY_ARRAY_UNIQUE_SYMBOL _cca_ARRAY_API
#include "numpy/ndarrayobject.h" 

// Macroes for less typing and better readability

#define I8(a,x) *(npy_uint8*)PyArray_GETPTR1(a,x)
#define X8(a,x,y) *(npy_uint8*)PyArray_GETPTR2(a,x,y)
#define Y8(a,i,x,y) *(npy_uint8*)PyArray_GETPTR3(a,i,x,y)
#define F8(a,i,x,y,c) *(npy_uint8*)PyArray_GETPTR4(a,i,x,y,c)
#define X32(a,x,y) *(npy_uint32*)PyArray_GETPTR2(a,x,y)
#define P8(a,x,y) (npy_uint8*)PyArray_GETPTR3(a,x,y,0)
#define Xint8(a,x,y) *(npy_int8*)PyArray_GETPTR2(a,x,y)

static PyObject *
CA2D(PyObject *dummy, PyObject *args)
{
  PyObject *pyo_rule=NULL, *pyo_lat=NULL, *pyo_img=NULL;
  int flip;
  if (!PyArg_ParseTuple(args, "OOOi", &pyo_rule, &pyo_lat, &pyo_img, &flip))
    return NULL;

  PyObject *rule=NULL, *lat=NULL, *img=NULL;

  // Access image array and set related variables
  rule = PyArray_FROM_OTF(pyo_rule, NPY_UINT8, NPY_INOUT_ARRAY);
  if (rule == NULL) {
    goto fail;
  }
  lat = PyArray_FROM_OTF(pyo_lat, NPY_UINT8, NPY_INOUT_ARRAY);
  if (lat == NULL) {
    goto fail;
  }
  img = PyArray_FROM_OTF(pyo_img, NPY_UINT8, NPY_INOUT_ARRAY);
  if (img == NULL) {
    goto fail;
  }

  npy_uint8 states[3];
  int state=0,newstate=0;
  int ndim=0;
  int flop=(flip+1)%2;
  npy_intp *dims=NULL;
  ndim = PyArray_NDIM(lat);
  dims = PyArray_DIMS(lat);

  int row=0, col=0, i=0, j=0;
  npy_uint8 *ptr=NULL;
  ptr = P8(img,0,0);
  for (row=0; row < dims[1]; row++) {
    for (i=0; i < 3; i++) {
      states[i] = 0;
      for (j=0; j < 3; j++) {
	states[i] += (Y8(lat, flip,
			 (row+dims[1]+i-1)%dims[1],
			 (col+dims[2]+j-2)%dims[2]) & 3) << ((2-j)*2) ;
      }
    }
    for (col=0; col < dims[2]; col++) {
      for (i=0; i < 3; i++) {
	states[i] = ((states[i] << 2) & 63) + 
	  (Y8(lat, flip, (row+dims[1]+i-1)%dims[1], (col+1)%dims[2]) & 3);
      }
      state = (states[0] << 13) | (states[1] << 7) | (states[2] << 1);
      X8(rule,1,(int)(state/8)) = 3 << (state%8);
      newstate = (X8(rule,0,(int)(state/8)) >> (state%8)) & 3;
      Y8(lat,flop,row,col) = newstate;
      *ptr = ((newstate&1) || ((newstate&2)>>1))*255; ptr++;
      *ptr = (newstate&2)*127; ptr++;
      *ptr = (newstate&1 && ((newstate&2)>>1))*255; ptr++;
    }
  }
  Py_INCREF(Py_None);
  return Py_None;
  
 fail:
  PyArray_XDECREF_ERR(rule);
  PyArray_XDECREF_ERR(lat);
  PyArray_XDECREF_ERR(img);
  return NULL;
}

//
// apply_rule function
//
// Implements the 1D binary CA rules of radius 1.
//

static PyObject *
apply_rule(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg=NULL, *arr=NULL;
  int rule=0;
  if (!PyArg_ParseTuple(args, "iO", &rule, &arg))
    return NULL;
  arr = PyArray_FROM_OTF(arg, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (arr == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument to array.");
    return NULL;
  }

  // We trust the caller to provide sane arguments
  // (the caller should be the python wrapping module)
  // so we do no argument checking here.

  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(arr);
  dims = PyArray_DIMS(arr);

  //For producing random data
  //char chs[(int)(dims[0]/8+1)];
  //char *ch;
  npy_ubyte *s, *t;             // source and target pointers
  int row, i;                   // row and column number
  int state;  /* This holds a binary number representing the state
                 of the neighbourhood we want to apply the rule to.
                 For example if the neighbourhood is 101, it's 5. */
  int right_end = dims[1] - 1;  // Precalculate for speed.
  int bottom_row = dims[0] - 1; // This too.
  s = PyArray_DATA(arr);        // Points to start of array
  t = s+dims[1];                // One row below s
  //ch = chs;

  // Now let's make a tight loop
  for (row=0; row < bottom_row; row++) {
    state = (*(s+right_end) << 1) + *s;
    for (i=0; i < right_end; i++) {
      state = ((state << 1) + *(++s)) & 7;
      *(t++) = ((rule >> state) & 1);
    }
    state = ((state << 1) + *((++s) - dims[1])) & 7;
    *t = ((rule >> state) & 1);

    /*
    // This is for producing an array of bytes in chs
    // based on the state of the last column of the 
    // array.
    *ch = *ch | (*t << (row % 8));
    if (row % 8 == 7) {
      ch++;
      *ch = 0;
    }
    */
    t++;
  }

  Py_DECREF(arr);

  // Return None
  Py_INCREF(Py_None);
  return Py_None;
  //return Py_BuildValue("s", chs);
};

//
// apply_code function
//
// Implements 1D radius 1 totalistic cellular automata
//

static PyObject *
apply_code(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg1=NULL, *code=NULL;
  PyObject *arg2=NULL, *arr=NULL;
  int radius=0;
  if (!PyArg_ParseTuple(args, "iOO", &radius, &arg1, &arg2))
    return NULL;
  code = PyArray_FROM_OTF(arg1, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (code == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument to array.");
    return NULL;
  }
  arr = PyArray_FROM_OTF(arg2, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (arr == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument to array.");
    goto fail;
  }

  // We trust the caller to provide sane arguments
  // (the caller should be the python wrapping module)
  // so we do no argument checking here.

  int ndim, w, h;
  npy_intp *dims;
  ndim = PyArray_NDIM(arr);
  dims = PyArray_DIMS(arr);
  w = dims[1];
  h = dims[0];

  int row, col;                   // row and column number
  int sum=0;
  int right_end = w - 1;          // Precalculate for speed.
  int bottom_row = h - 1;         // This too.

  // Now let's make a tight loop
  for (row=0; row < bottom_row; row++) {
    sum = 0;
    for (col=w-radius-1; col <= w+radius-1; col++) {
      sum += X8(arr, row, col % w);
    }
    for (col=0; col <= right_end; col++) {
      sum += X8(arr, row, (col+radius) % w) - 
	X8(arr, row, (col+w-radius-1) % w);
      X8(arr,row+1,col) = I8(code,sum);
    }
  }

  Py_DECREF(code);
  Py_DECREF(arr);

  Py_INCREF(Py_None);
  return Py_None;

 fail:
  PyArray_XDECREF_ERR(arr);
  PyArray_XDECREF_ERR(code);
  return NULL;
};


//
// kusch_markus function
//
// Implements a CA algorithm for producing may patterns found on
// seashells. It is due to Ingo Kusch and Mario Markus, thus the
// name. 
//

static PyObject *
kusch_markus(PyObject *dummy, PyObject *args, PyObject *keywds)
{
  // Start by parsing arguments
  PyObject *arg1=NULL, *arg2=NULL;
  PyObject *act=NULL;
  PyObject *inh=NULL;
  PyObject *tmpact=NULL;
  PyObject *tmpinh=NULL;
  int diff_r=17, nbh_r=1, w1=1, w2=1, seed=0;
  float m0=0, m1=0, d=0, p=0.002;

  // Parse arguments
  if (!PyArg_ParseTuple(args, "OOiiiiffffi", &arg1, &arg2, &diff_r,
			&nbh_r, &w1, &w2, &m0, &m1, &p, &d, &seed))
    return NULL;

  // Access or initialize act and inh arrays.
  act = PyArray_FROM_OTF(arg1, NPY_UINT8, NPY_INOUT_ARRAY);
  if (act == NULL) {
    goto fail;
  }
  inh = PyArray_FROM_OTF(arg2, NPY_UINT32, NPY_INOUT_ARRAY);
  if (inh == NULL) {
    goto fail;
  }

  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(act);
  dims = PyArray_DIMS(act);

  Py_ssize_t shape[2];
  shape[0] = 2; 
  shape[1] = dims[1];
  tmpact = PyArray_ZEROS(2, shape, NPY_UINT8, 0);
  if (tmpact == NULL) {
    goto fail;
  }
  tmpinh = PyArray_ZEROS(2, shape, NPY_UINT32, 0);
  if (tmpinh == NULL) {
    goto fail;
  }
  
  // This function is meant to be called by a wrapping function
  // in python that checks that the arguments are sane. Therefore
  // we don't do any further argument checking.

  int row, i, lastrow;          // row and column index
  int act_N;                    // Number of active cells within
                                // a radius of nbh_r
  int inh_N;                    // Sum of inhibitor within a
                                // radius of diff_r
  int inh_N_size = diff_r*2 + 1;// Number of elements in above sum
  
  // Initialize random number generator
  srand(seed);

  lastrow = dims[0]-1;
  d = 1 - d;
  for (row=1; row <= lastrow; row++) {
    // First we initialize the running sum act_N
    // Now for the loop itself, first pass
    for (i=0; i < dims[1]; i++) {
      // decay of inhibitor:
      if (X32(inh,row-1,i) != 0) {
	if (d != 1 && d != 0) {
	  X32(tmpinh,0,i) = (int)(d*X32(inh,row-1,i) + 0.5);
	} else {
	  X32(tmpinh,0,i) = X32(inh,row-1,i) - 1;
	}
      } else {
	X32(tmpinh,0,i) = 0;
      }
      // Activator production due to random fluctuations
      if (X8(act,row-1,i) == 0 && (rand() < (int)(p*RAND_MAX))) {
	X8(tmpact,0,i) = 1;
      } else {
	X8(tmpact,0,i) = X8(act,row-1,i);
      }
    }
    // Prepare running sum of activated cells
    act_N = X8(tmpact,0,0);
    for (i=1; i <= nbh_r; i++) {
      act_N += X8(tmpact,0,i) + X8(tmpact,0,dims[1]-i);
    }
    // Second pass
    for (i=0; i < dims[1]; i++) {
      // Production of inhibitor due to activator
      if (X8(tmpact,0,i) == 1) {
	X32(tmpinh,1,i) = X32(tmpinh,0,i) + w1;
	X8(tmpact,1,i) = 1;
      } else {
	X32(tmpinh,1,i) = X32(tmpinh,0,i);
	// Activator production from diffusion of activator
	if (act_N > (int)(m0 + m1*X32(tmpinh,1,i) + 0.5)) {
	  X8(tmpact,1,i) = 1;
	} else {
	  X8(tmpact,1,i) = X8(tmpact,0,i);
	}
      }
      // Update running sum
      act_N += X8(tmpact,0,(i+nbh_r+1)%dims[1]) - 
	X8(tmpact,0,(i+dims[1]-nbh_r)%dims[1]); 
      //printf("%d ", act_N);
    }
    // Prepare running sum of inhibitor amount
    inh_N = X32(tmpinh,1,0);
    for (i = 1; i <= diff_r; i++) {
      inh_N += X32(tmpinh,1,i) + X32(tmpinh,1,dims[1]-i);
    }
    // Third and final pass
    for (i=0; i < dims[1]; i++) { 
      // Diffusion of inhibitor
      X32(inh,row,i) = (int)(inh_N/inh_N_size + 0.5);
      // Update running sum of inhibitor amount
      inh_N += X32(tmpinh,1,(i+diff_r+1)%dims[1]) - 
	X32(tmpinh,1,(i+dims[1]-diff_r)%dims[1]);
      // Suppression of activator due to inhibitor
      if (X32(inh,row,i) >= w2) {
	X8(act,row,i) = 0;
      } else {
	X8(act,row,i) = X8(tmpact,1,i);
      }
    }
  }

  Py_DECREF(act);
  Py_DECREF(inh);
  Py_DECREF(tmpact);
  Py_DECREF(tmpinh);

  Py_INCREF(Py_None);
  return Py_None;

 fail:
  PyArray_XDECREF_ERR(act);
  PyArray_XDECREF_ERR(inh);
  PyArray_XDECREF_ERR(tmpact);
  PyArray_XDECREF_ERR(tmpinh);
  return NULL;
};


//
// hodgepodge machine
//
// The cyclic 2D CA called The Hodgepodge Machine
//

static PyObject *
hodgepodge(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg1=NULL, *arg2=NULL;
  PyObject *arr=NULL, *nbh=NULL;
  float k1=3, k2=3;
  int n=10, g=1;
  int source=0, target=0;

  // Parse arguments
  if (!PyArg_ParseTuple(args, "OOiiiff", &arg1, &arg2, &source, &n, &g, &k1, &k2))
    return NULL;

  // Access or initialize act and inh arrays.
  arr = PyArray_FROM_OTF(arg1, NPY_UINT8, NPY_INOUT_ARRAY);
  if (arr == NULL) {
    return NULL;
  }
  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(arr);
  dims = PyArray_DIMS(arr);

  nbh = PyArray_FROM_OTF(arg2, NPY_INT8, NPY_INOUT_ARRAY);
  if (nbh == NULL) {
    goto fail;
  }
  int nbhdim;
  npy_intp *nbhdims;
  nbhdim = PyArray_NDIM(nbh);
  nbhdims = PyArray_DIMS(nbh);

  int j, i, r;                        // array indices
  unsigned int N_ftd=0, N_ill=0, S=0;
  unsigned int state=0, newstate;

  //printf("%d %d %d %f %f", source, n, g, k1, k2);

  target = (source + 1) % 2;
  for (j=0; j < dims[1]; j++) {      
    for (i=0; i < dims[2]; i++) {
      N_ill = 0;
      N_ftd = 0;
      S = 0;
      for (r=0; r < nbhdims[0]; r++) {
	state = Y8(arr, source, (j+Xint8(nbh,r,0)+dims[1])%dims[1],	\
		   (i+Xint8(nbh,r,1)+dims[2])%dims[2]);
	if (state >= n) {
	  N_ill += 1;
	} else if (state > 0) {
	  N_ftd += 1;
	}
	S += state;
      }
      state = Y8(arr, source, j, i);
      if (state == 0) {
	newstate = (int)(N_ill/k1 + N_ftd/k2);
      } else if (state < n) {
	N_ftd += 1;
	newstate = (int)(g + S/N_ftd);
      } else {
	newstate = 0;
      }
      if (newstate > n) {
	newstate = n;
      }
      Y8(arr,target,j,i) = newstate;
      /*if (i+1 == dims[2] || j+1 == dims[1]) {
	Y8(arr,target,j,i) = 0;
	}*/
    }
  }


  Py_DECREF(arr);
  Py_DECREF(nbh);
  
  Py_INCREF(Py_None);
  return Py_None;
  
 fail:
  PyArray_XDECREF_ERR(arr);
  PyArray_XDECREF_ERR(nbh);
  return NULL;
  
};


static int funfun(int a, int b, int c) {
  return (int)((a + b/(256*4))/2);
}

static PyObject *
rgb(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg1=NULL, *arg2=NULL;
  PyObject *arr=NULL, *nbh=NULL;
  int source=0, target=0;

  // Parse arguments
  if (!PyArg_ParseTuple(args, "OOi", &arg1, &arg2, &source))
    return NULL;

  // Access or initialize act and inh arrays.
  arr = PyArray_FROM_OTF(arg1, NPY_UINT8, NPY_INOUT_ARRAY);
  if (arr == NULL) {
    return NULL;
  }
  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(arr);
  dims = PyArray_DIMS(arr);

  nbh = PyArray_FROM_OTF(arg2, NPY_INT8, NPY_INOUT_ARRAY);
  if (nbh == NULL) {
    goto fail;
  }
  int nbhdim;
  npy_intp *nbhdims;
  nbhdim = PyArray_NDIM(nbh);
  nbhdims = PyArray_DIMS(nbh);

  int j, i, c, r;                        // array indices
  unsigned int C[3];
  unsigned int d, dmax, rmax;

  target = (source + 1) % 2;
  for (j=0; j < dims[1]; j++) {
    for (i=0; i < dims[2]; i++) {
      for (c=0; c < dims[3]; c++) {
	C[c] = F8(arr, target, j, i, c);
      }
      dmax = 0; rmax = 0;
      for (r=0; r < nbhdims[0]; r++) {
	d = 0;
	for (c=0; c < dims[3]; c++) {
	  d += abs(F8(arr, target, (j+Xint8(nbh,r,0)+dims[1])%dims[1],	\
		      (i+Xint8(nbh,r,1)+dims[2])%dims[2],c) - C[c]);
	}
	if (d > dmax) {
	  dmax = d;
	  rmax = r;	    
	}
      }
      for (c=0; c < dims[3]; c++) {
	d = F8(arr, target, j, i, c);
	F8(arr, target, j, i, c) = F8(arr, target, (j+Xint8(nbh,rmax,0)+dims[1])%dims[1], (i+Xint8(nbh,rmax,1)+dims[2])%dims[2],c);
	F8(arr, target, (j+Xint8(nbh,rmax,0)+dims[1])%dims[1], (i+Xint8(nbh,rmax,1)+dims[2])%dims[2],c) = d;
      }
    }
  }


  Py_DECREF(arr);
  Py_DECREF(nbh);
  
  Py_INCREF(Py_None);
  return Py_None;
  
 fail:
  PyArray_XDECREF_ERR(arr);
  PyArray_XDECREF_ERR(nbh);
  return NULL;
  
};

static PyObject *
gol(PyObject *dummy, PyObject *args)
{
  //printf("hei...");

  // Start by parsing arguments
  PyObject *arg1=NULL, *arg2=NULL;
  PyObject *arr=NULL, *nbh=NULL;
  int source=0, target=0;

  // Parse arguments
  if (!PyArg_ParseTuple(args, "OOi", &arg1, &arg2, &source))
    return NULL;

  // Access or initialize act and inh arrays.
  arr = PyArray_FROM_OTF(arg1, NPY_UINT8, NPY_INOUT_ARRAY);
  if (arr == NULL) {
    return NULL;
  }
  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(arr);
  dims = PyArray_DIMS(arr);

  nbh = PyArray_FROM_OTF(arg2, NPY_INT8, NPY_INOUT_ARRAY);
  if (nbh == NULL) {
    goto fail;
  }
  int nbhdim;
  npy_intp *nbhdims;
  nbhdim = PyArray_NDIM(nbh);
  nbhdims = PyArray_DIMS(nbh);


  // This function is meant to be called by a wrapping function
  // in python that checks that the arguments are sane. Therefore
  // we don't do any further argument checking.

  int j, i, r;                        // array indices
  unsigned int S=0, state=0;

  target = (source + 1) % 2;
  for (j=0; j < dims[1]; j++) {      
    for (i=0; i < dims[2]; i++) {
      S = 0;
      for (r=0; r < nbhdims[0]; r++) {
	state = Y8(arr, source, (j+Xint8(nbh,r,0)+dims[1])%dims[1],	\
		   (i+Xint8(nbh,r,1)+dims[2])%dims[2]);
	S += state;
      }
      state = Y8(arr, source, j, i);
      if (S < 2) {
	Y8(arr,target,j,i) = 0;
      } else if (S == 2) {
	if (state == 1) {
	  Y8(arr,target,j,i) = 1;
	} else {
	  Y8(arr,target,j,i) = 0;
	}
      } else if (S == 3) {
	Y8(arr,target,j,i) = 1;
      } else {
	Y8(arr,target,j,i) = 0;
      }
    }
  }

  Py_DECREF(arr);
  Py_DECREF(nbh);
  
  Py_INCREF(Py_None);
  return Py_None;
  
 fail:
  PyArray_XDECREF_ERR(arr);
  PyArray_XDECREF_ERR(nbh);
  return NULL;
};

static PyObject *
latticegas(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg1=NULL;
  PyObject *arr=NULL;
  int source=0, target=0;

  // Parse arguments
  if (!PyArg_ParseTuple(args, "Oi", &arg1, &source))
    return NULL;

  // Access or initialize act and inh arrays.
  arr = PyArray_FROM_OTF(arg1, NPY_UINT8, NPY_INOUT_ARRAY);
  if (arr == NULL) {
    return NULL;
  }
  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(arr);
  dims = PyArray_DIMS(arr);

  int j, i;                        // array indices
  int a, b, c, d;                  // neighbourhood states

  target = (source + 1) % 2;
  for (j=source; j < dims[1]+source; j += 2) {      
    for (i=source; i < dims[2]+source; i += 2) {
      a = Y8(arr,source,j%dims[1],i%dims[2]);
      b = Y8(arr,source,j%dims[1],(i+1)%dims[2]);
      c = Y8(arr,source,(j+1)%dims[1],i%dims[2]);
      d = Y8(arr,source,(j+1)%dims[1],(i+1)%dims[2]);
      if (a == c && b == d && a != b) {
	Y8(arr,target,j%dims[1],i%dims[2]) = c;
	Y8(arr,target,j%dims[1],(i+1)%dims[2]) = d;
	Y8(arr,target,(j+1)%dims[1],i%dims[2]) = a;
	Y8(arr,target,(j+1)%dims[1],(i+1)%dims[2]) = b;
      } else {
	Y8(arr,target,j%dims[1],i%dims[2]) = d;
	Y8(arr,target,j%dims[1],(i+1)%dims[2]) = c;
	Y8(arr,target,(j+1)%dims[1],i%dims[2]) = b;
	Y8(arr,target,(j+1)%dims[1],(i+1)%dims[2]) = a;
      }
    }
  }

  Py_DECREF(arr);
  
  Py_INCREF(Py_None);
  return Py_None;
};

//
// two_dynamics
//
// Implements the 1D binary CA rules of radius 1 with 2-dynamics
//

static PyObject *
two_dynamics(PyObject *dummy, PyObject *args)
{

  // Start by parsing arguments
  PyObject *aL=NULL, *aN=NULL;
  PyObject *L=NULL, *N=NULL, *R=NULL, *F=NULL;

  if (!PyArg_ParseTuple(args, "OOOO", &aL, &aN, &R, &F))
    return NULL;

  L = PyArray_FROM_OTF(aL, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (L == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert lattice to array.");
    goto fail;
  }
  int Ldim;
  npy_intp *Ldims;
  Ldim = PyArray_NDIM(L);
  Ldims = PyArray_DIMS(L);

  N = PyArray_FROM_OTF(aN, NPY_BYTE, NPY_INOUT_ARRAY);
  if (N == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert neighbourhood set to array.");
    goto fail;
  }
  int Ndim;
  npy_intp *Ndims;
  Ndim = PyArray_NDIM(N);
  Ndims = PyArray_DIMS(N);

  // S holds the states of all the cells in the neighbourhood, in order
  Py_ssize_t Sdims[1];
  Sdims[0] = Ndims[1]+1;
  PyObject *S=NULL;
  S = PyArray_ZEROS(1, Sdims, NPY_INT8, 0);
  if (S == NULL) {
    goto fail;
  }

  int x, t, n;
  int nbh, nbh_max, dx, s, sum;
  int rule;
  PyObject *result, *arguments;

  for (t=0; t < Ldims[0]-1; t++) {
    for (x=0; x < Ldims[1]; x++) {
      nbh = Y8(L, t, x, 1);
      rule = Y8(L, t, x, 2);
      // The 0th column of every row is the index of the row that
      // follows the last item of that row
      nbh_max = Xint8(N, nbh, 0);
      I8(S, 0) = nbh_max;
      sum = 0;
      for (n=1; n < nbh_max; n++) {
	dx = Xint8(N, nbh, n);
	s = Y8(L, t, (x+dx+Ldims[1])%Ldims[1], 0);
	sum += s;
	I8(S, n) = s;
      }
      I8(S, nbh_max) = sum;
      // Now S[0] is the index of the sum, so S[S[0]] is the sum. 
      // S[1] should be the state of the center cell, as long as
      // N is correctly formatted.

      result = PyEval_CallObject(PyList_GET_ITEM(R, rule), 
				 Py_BuildValue("(O)", S));
      if (result == NULL)
	goto fail;
      //printf("%d ", PyInt_AS_LONG(result));
      //fflush(stdout);
      Y8(L, t+1, x, 0) = PyInt_AS_LONG(result);
      Py_DECREF(result);
      arguments = Py_BuildValue("(OiOiO)", S, nbh, N, rule, R);
      result = PyEval_CallObject(F, arguments);
      Py_DECREF(arguments);
      if (result == NULL)
	goto fail;
      Y8(L, t+1, x, 1) = PyInt_AS_LONG(PyTuple_GET_ITEM(result, 0));
      Y8(L, t+1, x, 2) = PyInt_AS_LONG(PyTuple_GET_ITEM(result, 1));
      Py_DECREF(result);
    }
  }
  /*
    for (t=0; t < 2; t++) {
    for (x=0; x < 20; x++) {
      printf("%d ", Y8(L, t, x, 0));
    }
    printf("\n");
    }*/
  Py_DECREF(L);
  Py_DECREF(N);
  Py_DECREF(S);

  // Return None
  Py_INCREF(Py_None);
  return Py_None;

 fail:
  PyArray_XDECREF_ERR(L);
  PyArray_XDECREF_ERR(N);
  PyArray_XDECREF_ERR(S);
  Py_XDECREF(arguments);
  return NULL;
};


static PyObject *
paint_img(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg1=NULL, *arg2=NULL, *arg3=NULL;
  PyObject *pat=NULL, *img=NULL, *pal=NULL;
  if (!PyArg_ParseTuple(args, "OOO", &arg1, &arg2, &arg3))
    return NULL;
  pat = PyArray_FROM_OTF(arg1, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (pat == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument 1 to array.");
    return NULL;
  }
  img = PyArray_FROM_OTF(arg2, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (img == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument 2 to array.");
    goto fail;
  }
  pal = PyArray_FROM_OTF(arg3, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (pal == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument 3 to array.");
    goto fail;
  }

  // We trust the caller to provide sane arguments
  // (the caller should be the python wrapping module)
  // so we do no dimension checking here.
  // pat should be (y,x) and img should be (y,x,3)
  // pal is the palette of pat, so it's (n,3), where
  // n is the highest value of the array pat.

  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(img);
  dims = PyArray_DIMS(img);

  int y=0,x=0,i=0;
  npy_ubyte *img_p, *pal_p;
  img_p = (npy_uint8*)PyArray_GETPTR3(img,0,0,0);
  for (y=0; y < dims[0]; y++) {
    for (x=0; x < dims[1]; x++) {
      pal_p = (npy_uint8*)PyArray_GETPTR2(pal,X8(pat,y,x),0);
      for (i=0; i < 3; i++) {
	*(img_p++) = *(pal_p++);
      }
    }
  }

  Py_DECREF(pat);
  Py_DECREF(img);
  Py_DECREF(pal);

  Py_INCREF(Py_None);
  return Py_None;

 fail:
  PyArray_XDECREF_ERR(pat);
  PyArray_XDECREF_ERR(img);
  PyArray_XDECREF_ERR(pal);
  return NULL;
};


static PyObject *
normalize_img(PyObject *dummy, PyObject *args)
{
  // Start by parsing arguments
  PyObject *arg1=NULL, *arg2=NULL;
  PyObject *pat=NULL, *max=NULL;
  if (!PyArg_ParseTuple(args, "OO", &arg1, &arg2))
    return NULL;
  pat = PyArray_FROM_OTF(arg1, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (pat == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument 1 to array.");
    return NULL;
  }
  max = PyArray_FROM_OTF(arg2, NPY_UBYTE, NPY_INOUT_ARRAY);
  if (max == NULL) {
    PyErr_SetString(PyExc_TypeError, "Could not convert argument 2 to array.");
    goto fail;
  }

  int ndim;
  npy_intp *dims;
  ndim = PyArray_NDIM(pat);
  dims = PyArray_DIMS(pat);

  int y=0,x=0,i=0,t=0;
  npy_ubyte *pat_p;
  pat_p = (npy_uint8*)PyArray_GETPTR3(pat,0,0,0);
  for (y=0; y < dims[0]; y++) {
    for (x=0; x < dims[1]; x++) {
      for (i=0; i < 3; i++) {
	t = *(pat_p)*255/I8(max,i);
	*(pat_p++) = t;
      }
    }
  }

  Py_DECREF(pat);
  Py_DECREF(max);

  Py_INCREF(Py_None);
  return Py_None;

 fail:
  PyArray_XDECREF_ERR(pat);
  PyArray_XDECREF_ERR(max);
  return NULL;
};


//
// The below is standard python extension stuff
//

static PyMethodDef _ccamethods[] = {
    {"apply_rule",  apply_rule, METH_VARARGS,
     "Apply a 1-radius binary rule to an array."},
    {"apply_code",  apply_code, METH_VARARGS,
     "Apply a 1-radius totalistic rule to an array."},
    {"kusch_markus",  (PyCFunction)kusch_markus, 
     METH_VARARGS|METH_KEYWORDS,
     "A CA model due to Ingo Kusch and Mario Markus."},
    {"CA2D", CA2D, METH_VARARGS,
     "Apply a Moore neighbourhood 2D binary rule to an array."},
    {"hodgepodge",  hodgepodge, METH_VARARGS,
     "The Hodgepodge Machine."},
    {"rgb",  rgb, METH_VARARGS,
     "Experimental CA with 3 coupled 256-state lattices."},
    {"gol",  gol, METH_VARARGS,
     "The Game of Life."},
    {"latticegas",  latticegas, METH_VARARGS,
     "Lattice Gas"},
    {"two_dynamics", two_dynamics, METH_VARARGS,
     "A 1-CA with 2-dynamics"},
    {"paint_img",  paint_img, METH_VARARGS,
     "Draw a CA pattern to an image-like array."},
    {"normalize_img",  normalize_img, METH_VARARGS,
     "Scale color component values to cover values 0-255."},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC
init_cca(void)
{
  (void)Py_InitModule("_cca", _ccamethods);
  import_array();
};

int
main(int argc, char *argv[])
{
    Py_SetProgramName(argv[0]);
    Py_Initialize();
    init_cca();
    return 0;
};
