#!/usr/bin/env python

#
# _cca.so wrapper module
# Copyright (C) 2009 Lauritz Thaulow
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import _cca
import numpy as np
from types import IntType, FloatType, StringType, UnicodeType
from random import uniform, randint
from colorsys import hsv_to_rgb
from math import sqrt
from time import time

__all__ = ["SMALL_STEP", "MEDIUM_STEP", "LARGE_STEP", 
           "KuschMarkus", "CA1D_bin", "CA1D_tot", "Hodgepodge",
           "RGB", "GOL", "LatticeGas", "TwoDynamics"]

SMALL_STEP = 0
MEDIUM_STEP = 1
LARGE_STEP = 2

def percent_stepper(obj, param, step, dir):
    mult = 0.1
    if step == SMALL_STEP:
        return dir/abs(dir)
    elif step == LARGE_STEP:
        mult = 1
    ret = mult*param.__get__(obj)
    if dir < 0:
        ret /= (1+step)
    return ret*(dir/abs(dir))


def small_decimal_stepper(obj, param, step, dir):
    mult = 1
    if step == LARGE_STEP:
        mult = 10
    return mult*(dir/abs(dir))

def sqrt_stepper(obj, param, step, dir):
    mult = 1
    rang = param.get_max(obj) - param.get_min(obj)
    if step == MEDIUM_STEP:
        mult = sqrt(sqrt(rang))
    elif step == LARGE_STEP:
        mult = sqrt(rang)
    return int(mult)*(dir/abs(dir))


class Param(object):
    def __init__(self, name, default=0, **kwargs):
        #super(Param, self).__init__()
        self.name = "_" + name
        self.default = default
        self.apply_args(kwargs)

    def apply_args(self, kwargs):
        self.hidden = kwargs.get("hidden", False)
        self.no_compare = kwargs.get("no_compare", False)
        self.on_change = kwargs.get("on_change", [])
        
    def reset(self, obj):
        self.__set__(obj, self.default)

    def check(self, obj):
        self.__set__(obj, getattr(obj, self.name))

    def __get__(self, obj, type=None):
        try:
            val = getattr(obj, self.name)
        except AttributeError:
            val = self.default
        return val

    def __set__(self, obj, val):
        old_v = self.__get__(obj)
        val = self.constrain(obj, self.parse(val))
        setattr(obj, self.name, val)
        if self.no_compare or val != old_v:
            for code in self.on_change:
                if type(code) == StringType:
                    eval(code)
                else:
                    code(vars())
    
    def call_on_change(fn_or_code):
        self.on_change.append(fn_or_code)

    def parse(self, val):
        return val

    def constrain(self, obj, val):
        return val

    def __repr__(self):
        return "|%s|" % repr(self._value)

    def has_changed(self, obj, reset):
        # Prototype
        return False

    def getstring(self, obj):
        val = getattr(obj, self.name)
        if type(val) in (IntType, FloatType):
            return repr(val)
        return val


class TrackedParam(Param):
    def __init__(self, *args, **kwargs):
        super(TrackedParam, self).__init__(*args, **kwargs) 
        self.changed = self.name + "_changed"
        self.on_change.append("setattr(obj, self.changed, True)")

    def has_changed(self, obj, reset=True):
        try:
            ret = getattr(obj, self.changed)
        except AttributeError:
            ret = None        
        if reset and ret:
            setattr(obj, self.changed, False)
        return ret


class OnlyCertainNumbers(TrackedParam):
    def apply_args(self, kwargs):
        super(OnlyCertainNumbers, self).apply_args(kwargs)
        self.stepper = kwargs.get("stepper", small_decimal_stepper)

    def step(self, obj, direction, step=MEDIUM_STEP):
        direction = direction/abs(direction)
        step = self.stepper(obj, self, step, direction)
        old_v = self.__get__(obj)
        new_v = self.constrain(obj, self.parse(old_v + step))
        if new_v == old_v:
            new_v += direction
        self.__set__(obj, new_v)

    def tmp(self):
        if hasattr(self, "display"):
            self.display.ChangeValue(repr(self._value))
            self.display.force_redraw()


class OnlyIntegers(OnlyCertainNumbers): 
    def parse(self, val):
        if type(val) in (StringType, UnicodeType) and val[-1] == "L":
            val = val[:-1]
        return int(val)


class OnlyIntegerAbove(OnlyIntegers):
    def apply_args(self, kwargs):
        super(OnlyIntegerAbove, self).apply_args(kwargs)
        self.min = kwargs.get("min", 0)

    def get_min(self, obj):
        if type(self.min) == StringType:
            return eval(self.min)
        return self.min

    def constrain(self, obj, value):
        min = self.get_min(obj)
        if value < min:
            value = min
        return super(OnlyIntegerAbove, self).constrain(obj, value)


class OnlyIntegerBelow(OnlyIntegers):
    def apply_args(self, kwargs):
        super(OnlyIntegerBelow, self).apply_args(kwargs)
        self.max = kwargs.get("max", 1)

    def get_max(self, obj):
        if type(self.max) == StringType:
            return eval(self.max)
        return self.max
        
    def constrain(self, obj, value):
        max = self.get_max(obj)
        if value > max:
            value = max
        return super(OnlyIntegerBelow, self).constrain(obj, value)


class OnlyIntegerRange(OnlyIntegerBelow, OnlyIntegerAbove):
    def randomize(self, obj):
        self.__set__(obj, uniform(self.get_min(obj), self.get_max(obj)))
        print self.__get__(obj)



class OnlyFloats(OnlyCertainNumbers):
    def parse(self, value):
        return float(value)


class OnlyFloatAbove(OnlyFloats, OnlyIntegerAbove):
    pass

class OnlyFloatBelow(OnlyFloats, OnlyIntegerBelow):
    pass

class FloatRange(OnlyFloats, OnlyIntegerRange):
    pass


class CA1D_state(Param):
    def __init__(self, *args, **kwargs):
        super(CA1D_state, self).__init__(*args, **kwargs)
        self.no_compare = True

    def __set__(self, obj, val):
        if val.base != None:
            val = val.copy()
        super(CA1D_state, self).__set__(obj, val)

    def parse(self, value):
        return value


class CA1D_init(TrackedParam):
    def __init__(self, *args, **kwargs):
        super(CA1D_init, self).__init__(*args, **kwargs)
        self.no_compare = True

    def getstring(self, obj):
        return str(getattr(obj, self.name))



class ClassWithParams(type):  # Metaclass
    def __init__(cls, name, bases, dct):
        super(ClassWithParams, cls).__init__(name, bases, dct)
        d = {}
        for c in cls.__mro__:
            for (k, v) in c.__dict__.items():
                if issubclass(v.__class__, Param):
                    d[k] = v
        cls.param = d
    

class CA1D(object):
    __metaclass__ = ClassWithParams

    width = OnlyIntegerAbove(
        "width", 800, min=5, stepper=percent_stepper)
    steps = OnlyIntegerAbove(
        "steps", 101, min=3, hidden=True)
    state = CA1D_state(
        "state", default=np.zeros(()), hidden=True, on_change=[])
    #init = CA1D_init("init", default="R")
    
    def __init__(self):
        for i in self:
            self[i].reset(self)
        self.arr = None
        self.img = None
        self.set_palette()

    def set_palette(self):
        pass

    def has_changed(self, reset=False):
        changed = False
        for (k, v) in self.__class__.param.items():
            if v.has_changed(self, reset):
                changed = True
        return changed

    def check_param(self, name):
        self.__class__.param[name].check(self)

    def new_arrs(self):
        (w, h) = (self.width, self.steps)
        self.arr = np.zeros((h, w), dtype="uint8")
        self.img = np.zeros((h, w, 3), dtype="uint8")

    def resize_arrs(self):
        (w, h) = (self.width, self.steps)
        self.arr.resize((h, w))
        self.img.resize((h, w, 3))
        
    def init_arrs(self, reset=False):
        if self.arr == None:
            self.new_arrs()
        else:
            try:
                self.resize_arrs()
            except:
                self.new_arrs()
        ok = True
        if not hasattr(self.state, "shape"):
            ok = False
        if not ok or reset == True:
            self.reset()

    def reset(self):
        state = self.arr[0,:].copy()
        state[:] = np.random.randint(0, 2, self.arr.shape[1:])
        #if self.init.lower() == "r":
        #    state[:] = np.random.randint(0, 2, self.arr.shape[1:])
        #else:
        #    pat = [int(x) for x in self.init.split()]
        #    start = self.width/2-len(pat)/2
        #    end = start + len(pat)
        #    state[start:end] = pat
        self.state = state

    def get_img(self):
        if self.img == None:
            raise RuntimeError("No image produced yet, call run() first!")
        _cca.paint_img(self.arr, self.img, self.pal)
        return self.img

    def run(self):
        # Prototype
        shape = (self.steps, self.width)
        if self.arr == None or self.arr.shape[:len(shape)] != shape:
            # Now we init the arrays.
            self.init_arrs()
        self.has_changed(reset=True)

    def get_max_radius(self):
        return self.width/2-1
    max_radius = property(get_max_radius)

    def __getitem__(self, key):
        return self.__class__.param[key]

    def __iter__(self):
        return iter(self.__class__.param)


class CA1D_bin(CA1D):

    rule = OnlyIntegerRange(
        "rule", 30, min=0, max=255, stepper=small_decimal_stepper)

    def set_palette(self):
        self.pal = np.zeros((2,3), dtype="uint8")
        self.pal[0,:] = 255
        self.pal[1,:] = 0

    def run(self):
        super(CA1D_bin, self).run()
        self.arr[0,:] = self.state
        _cca.apply_rule(self._rule, self.arr)
        self.state = self.arr[-1,:]


class CA1D_tot(CA1D):

    radius = OnlyIntegerRange(
        "radius", 1, min=0, max="obj.max_radius", 
        stepper=small_decimal_stepper, 
        on_change=["obj.check_param('code')"])

    states = OnlyIntegerAbove(
        "states", 3, min=2, stepper=small_decimal_stepper,
        on_change=["obj.check_param('code')"])

    code = OnlyIntegerRange(
        "code", 1599, min=0, max="obj.max_code", stepper=sqrt_stepper)

    def set_palette(self):
        self.pal = np.zeros((self.states,3), dtype="uint8")
        self.pal[0,:] = 255
        self.pal[-1,:] = 0
        states = self.states - 2
        huesteps = 32
        layers = int(states/huesteps)
        if layers <= 1:
            layers = 1
        i = 1
        val_step = 1.0/(layers + 1)
        val = val_step
        hue_step = 1.0/huesteps
        while i < states+1:
            if i % layers == 0:
                val += val_step
            c = hsv_to_rgb(float((i+4) % huesteps)/huesteps, 1, val)
            for (j, v) in enumerate(c):
                self.pal[i,j] = int(v*255)
            i += 1
        
    def run(self):
        if self.radius*2+1 > self.width:
            raise RuntimeError("Neighbourhood is larger than lattice.")
        self.set_palette()
        super(CA1D_tot, self).run()
        self.arr[0,:] = self.state
        _cca.apply_code(self._radius, self.code_array(self.code), self.arr)
        self.state = self.arr[-1,:]

    def get_max_code(self):
        states = self.states
        radius = self.radius
        nbh_n = (states-1)*(radius*2+1)+1
        return states**nbh_n - 1
    max_code = property(get_max_code)
        
    def code_array(self, code):
        states = self.states
        radius = self.radius
        nbh_n = (states-1)*(radius*2+1)+1
        code_arr = np.zeros((nbh_n,), dtype="uint8")
        n = 0
        while code != 0:
            rem = code % states
            code_arr[n] = rem
            code -= rem
            code /= states
            n += 1
        return code_arr


class KuschMarkusState(object):
    def __init__(self, arr, inh, seed, index=0):
        if arr != None:
            self.arr = arr[index,:].copy()
        if inh != None:
            self.inh = inh[index,:].copy()
        if seed != None:
            self.seed = int(seed)

    def copy(self):
        s = KuschMarkusState(None, None, None)
        s.arr = self.arr.copy()
        s.inh = self.inh.copy()
        s.seed = self.seed
        return s

    def get_shape(self):
        return self.arr.shape
    shape = property(get_shape)

    def get_base(self):
        return self.arr.base
    base = property(get_base)

        
class KuschMarkus(CA1D):

    diff_r = OnlyIntegerRange(
        "diff_r", 17, min=1, max="obj.max_radius", stepper=percent_stepper)

    nbh_r = OnlyIntegerRange(
        "nbh_r", 1, min=1, max="obj.max_radius", stepper=percent_stepper)
    
    w1 = OnlyIntegerAbove(
        "w1", 1, min=0, stepper=percent_stepper)

    w2 = OnlyIntegerAbove(
        "w2", 1, min=0, stepper=percent_stepper)

    m0 = OnlyFloatAbove(
        "m0", 0, min=0, stepper=percent_stepper)

    m1 = OnlyFloatAbove(
        "m1", 0, min=0, stepper=percent_stepper)

    d = FloatRange(
        "d", 0, min=0, max=1, stepper=percent_stepper)
    
    p = FloatRange(
        "p", 1.0/500, min=0, max=1, stepper=percent_stepper)

    def new_arrs(self):
        super(KuschMarkus, self).new_arrs()
        self.inh = np.zeros((self.steps, self.width), dtype="uint32")

    def resize_arrs(self):
        super(KuschMarkus, self).resize_arrs()
        self.inh.resize((self.steps, self.width))

    def set_palette(self):
        self.pal = np.zeros((2,3), dtype="uint8")
        self.pal[0,:] = 255
        self.pal[1,:] = 0

    def reset(self):
        s = KuschMarkusState(self.arr, self.inh, time())
        s.arr[:] = 0
        s.inh[:] = 0
        s.arr[self.width/2] = 1
        self.state = s

    def get_img(self):
        if self.img == None:
            raise RuntimeError("No image produced yet, call run() first!")
        _cca.paint_img(self.arr, self.img, self.pal)
        return self.img
        
    def run(self):
        super(KuschMarkus, self).run()
        s = self.state
        self.arr[0,:] = s.arr
        self.inh[0,:] = s.inh
        _cca.kusch_markus(self.arr, self.inh, self._diff_r, self._nbh_r,
                          self._w1, self._w2, self._m0, self._m1, self._p,
                          self._d, self._state.seed)
        self.state = KuschMarkusState(self.arr, self.inh, time(), -1)

def rule_factory(r):
    def rule(nbh):
        return (r >> ((nbh[1]<<2)+(nbh[2]<<1)+nbh[3])) & 1
    return rule

def quo(*args):
    return (0,0)

def testphi(*args):
    n = args[3]
    sum = args[0][-1]
    new = n
    if sum == 0:
        new = 1
    if sum == 3:
        new = 0
    return (0, new)

class TwoDynamics(CA1D):
    def __init__(self):
        super(TwoDynamics, self).__init__()
        self.N = np.zeros((1,4), dtype="int8")
        self.N[0,:] = [4, -1, 0, 1]
        self.R = [rule_factory(32), rule_factory(251)]
        self.F = testphi

    def new_arrs(self):
        (s, w) = (self.steps, self.width)
        self.arr = np.zeros((s, w, 3), dtype="uint8")
        self.img = np.zeros((s, w, 3), dtype="uint8")

    def resize_arrs(self):
        (s, w) = (self.steps, self.width)
        self.arr.resize((s,w,3))
        self.img.resize((s,w,3))
        self.reset()

    def set_palette(self):
        self.pal = np.zeros((2,3), dtype="uint8")
        self.pal[0,:] = 0
        self.pal[1,:] = 255

    def reset(self):
        self.arr[0,:,0] = np.random.randint(0, 2, (self.width))
        #self.arr[0,:,0] = 1
        self.arr[0,:,1:2] = 0
        self.state = self.arr[0,...]

    def get_img(self):
        if self.img == None:
            raise RuntimeError("No image produced yet, call run() first!")
        self.img = self.arr[...]
        maxes = np.zeros((3,), dtype="uint8")
        maxes[...] = (1,1,1)
        _cca.normalize_img(self.img, maxes)
        return self.img
        
    def run(self):
        super(TwoDynamics, self).run()
        self.arr[0,...] = self.state
        _cca.two_dynamics(self.arr, self.N, self.R, self.F)
        self.state = self.arr[-1,...]
        


class CA2D(CA1D):
    height = OnlyIntegerAbove("height", 240)

    def __init__(self):
        super(CA2D, self).__init__()
        self.steps = 3
        self.counter = 0

    def new_arrs(self):
        (s, w, h) = (self.steps, self.width, self.height)
        self.arr = np.zeros((2, h, w), dtype="uint8")
        self.img = np.zeros((h, w, 3), dtype="uint8")

    def resize_arrs(self):
        (s, w, h) = (self.steps, self.width, self.height)
        self.arr.resize((2, h, w))
        self.img.resize((h, w, 3))

    def init_arrs(self, reset=False):
        if self.arr == None:
            self.new_arrs()
        else:
            try:
                self.resize_arrs()
            except:
                self.new_arrs()
        if reset == True:
            self.reset()

    def get_img(self, step):
        print "getting img of step ", step
        if self.img == None:
            raise RuntimeError("No image produced yet, call run() first!")
        _cca.paint_img(self.arr[int(step)%2,...], self.img, self.pal)
        return self.img


class GOL(CA2D):
    def __init__(self, *args, **kwargs):
        super(GOL, self).__init__(*args, **kwargs)
        self.neighbourhood = np.zeros((8,2), dtype="int8")
        n = [(i,j) for i in range(-1,2) for j in range(-1,2)]
        n.pop(n.index((0,0)))
        self.neighbourhood[:] = n
        self.set_palette()
    
    def set_palette(self):
        self.pal = np.zeros((2,3), dtype="uint8")
        self.pal[0,:] = 255
        self.pal[-1,:] = 0
        
    def run(self):
        super(GOL, self).run()
        #self.arr[0,...] = self.state
        src = int(self.counter) % 2
        _cca.gol(self.arr, self.neighbourhood, src)
        print "Running, count %d" % self.counter
        self.counter += 1
        #self.state = self.arr[-1,...]

    def reset(self):
        (w, h) = (self.width, self.height)
        self.arr[0,...] = 0
        self.arr[0,h/2+1, w/2-3:w/2+4] = [1, 1, 0, 0, 1, 1, 1]
        self.arr[0,h/2, w/2] = 1
        self.arr[0,h/2-1,w/2-2] = 1
        self.counter = 0

class RGB(CA2D):
    def __init__(self, *args, **kwargs):
        super(RGB, self).__init__(*args, **kwargs)
        #self.neighbourhood = np.zeros((4,2), dtype="int8")
        #self.neighbourhood[:] = [(-1,0),(1,0),(0,-1),(0,1)]
        self.neighbourhood = np.zeros((8,2), dtype="int8")
        n = [(i,j) for i in range(-1,2) for j in range(-1,2)]
        n.pop(n.index((0,0)))
        self.neighbourhood[:] = n

    def new_arrs(self):
        (s, w, h) = (self.steps, self.width, self.height)
        self.arr = np.zeros((2, h, w, 3), dtype="uint8")

    def resize_arrs(self):
        (s, w, h) = (self.steps, self.width, self.height)
        self.arr.resize((2, h, w, 3))
        
    def run(self):
        super(RGB, self).run()
        src = int(self.counter) % 2
        self.arr[(src + 1) % 2, ...] = self.arr[src, ...]
        _cca.rgb(self.arr, self.neighbourhood, src)
        print "Running, count %d" % self.counter
        self.counter += 1

    def reset(self):
        (w, h) = (self.width, self.height)
        self.arr[0,...] = np.random.randint(0, 256, (h, w, 3))
        self.counter = 0

    def get_img(self, step):
        if self.arr == None:
            raise RuntimeError("No image produced yet, call run() first!")
        return self.arr[int(self.counter % 2),...]

class LatticeGas(CA2D):
    def __init__(self, *args, **kwargs):
        super(LatticeGas, self).__init__(*args, **kwargs)
        self.set_palette()
    
    def set_palette(self):
        self.pal = np.zeros((2,3), dtype="uint8")
        self.pal[0,:] = 255
        self.pal[-1,:] = 0
        
    def run(self):
        super(LatticeGas, self).run()
        src = int(self.counter) % 2
        _cca.latticegas(self.arr, src)
        print "Running count %d" % self.counter
        self.counter += 1

    def reset(self):
        (w, h) = (self.width, self.height)
        self.arr[0,0:h/2,0:w/2] = np.random.randint(0, 2, (h/2, w/2))
        self.counter = 0


class Hodgepodge(CA2D):
    k1 = OnlyFloatAbove("k1", 2, min=0.00001, stepper=percent_stepper)
    k2 = OnlyFloatAbove("k2", 3.5, min=0.00001, stepper=percent_stepper)
    g = OnlyIntegerRange("g", 19, min=1, max="getattr(obj, 'n')",
                   stepper=small_decimal_stepper)
    n = OnlyIntegerAbove("n", 100, min=3, stepper=small_decimal_stepper)

    def __init__(self, *args, **kwargs):
        super(Hodgepodge, self).__init__(*args, **kwargs)
        self.neighbourhood = np.zeros((8,2), dtype="int8")
        n = [(i,j) for i in range(-1,2) for j in range(-1,2)]
        n.pop(n.index((0,0)))
        self.neighbourhood[:] = n
        self.set_palette()
    
    def set_palette(self):
        self.pal = np.zeros((self.n+1,3), dtype="uint8")
        for i in range(self.n+1):
            for j in range(3):
                self.pal[i,j] = 50+150*i/(self.n+1) 
        self.pal[0,...] = 0
        self.pal[self.n,...] = 255
        
    def run(self):
        super(Hodgepodge, self).run()
        #self.arr[0,...] = self.state
        #n = self.arr[self.counter % 2,self.height-1,self.width-1] 
        #newn = n
        #while newn >= n:
        src = self.counter % 2
        _cca.hodgepodge(self.arr, self.neighbourhood, src, self._n, self._g, self._k1, self._k2)
        self.counter += 1
        #    n = newn
        #    newn = self.arr[self.counter % 2,self.height-1,self.width-1]
        #self.state = self.arr[-1,...]

    def reset(self):
        self.arr[0,...] = np.random.randint(0, self.n+1, self.arr.shape[1:])
        #self.state = self.arr[0,...]

