/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at http://www.qtsoftware.com/contact.
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QtGui>
#include <iostream>
#include <math.h>

#include "renderthread.h"

#include "rules.h"

// Dimension of lattice
const unsigned int LAT_W = 200;
const unsigned int LAT_H = 200;

// Distance between lattices
const unsigned int SEP = 10;

// Time steps between display update
const unsigned int STEP = 10;

// The number of time steps of constant number of changes
// required before concluding that a pattern has stabilized
const unsigned int UNCHANGED_QUARANTINE = 400;

typedef std::vector<CA1D_ptr> CA1D_vec;

RenderThread::RenderThread(QObject *parent)
    : QThread(parent), state(State_ptr(new State(Coord(LAT_W,LAT_H), Nbhs_ptr(), Rules())))
{
    restart = false;
    abort = false;
    ca = NULL;
}

//1 12407 13703 3066052511 3066053847
RenderThread::~RenderThread()
{
    mutex.lock();
    abort = true;
    condition.wakeOne();
    mutex.unlock();

    wait();
}

void RenderThread::render()
{
    QMutexLocker locker(&mutex);

    if (!isRunning()) {
        start(LowPriority);
    } else {
        restart = true;
        condition.wakeOne();
    }
}

void RenderThread::run()
{
    const unsigned int LAT_W_SEP = SEP + LAT_W;

    forever {
        state->clear();
        state->random(30, StateIdx);
        CA1D_vec cas;
        cas.push_back(initCA());
        cas.push_back(initHOCA());

        QRect rect = QRect(0, 0, LAT_W_SEP*cas.size() + SEP, LAT_H + 2*SEP);
        QImage image = QImage(rect.size(), QImage::Format_RGB32);
        image.fill(QColor(255,255,255).rgb());

        int cnt = 0;
        std::cout << std::endl << std::flush;

        forever {
            cnt++;
            if (restart)
                break;
            if (abort)
                return;

            unsigned int col = 0;
            CA1D_vec::iterator ca_i, ca_e;
            ca_e = cas.end();
            unsigned int unchanged_cas = 0;
            for (ca_i = cas.begin(); ca_i != ca_e; ca_i++) {
                col++;
                CA1D_ptr ca = *ca_i;
                State_ptr st = ca->get();
                if (static_cast<int>(ca->last_changes/4) == static_cast<int>(st->changes)/4) {
                    ca->unchanged_for++;
                } else {
                    ca->last_changes = st->changes;
                    ca->unchanged_for = 0;
                    ca->last_change_at = cnt;
                }
                if (ca->unchanged_for > UNCHANGED_QUARANTINE) {
                    unchanged_cas++;
                }
                //std::cout << " " << st->changes << " " << st->live_cells;

                const unsigned char *a;
                unsigned char r;
                for (int y = 0; y < LAT_H; ++y) {
                    uint *scanLine =
                            reinterpret_cast<uint *>(image.scanLine(y+SEP));
                    a = st->scanLine(y);
                    scanLine += SEP + col*LAT_W_SEP;
                    unsigned int stride = st->x_stride;
                    for (int x = 0; x != LAT_W; x++) {
                        //*scanLine++ = qRgb(252,123,0);
                        if (*(a + x*stride) == 0) {
                            r = 255;
                        } else {
                            r = 0;
                        }
                        //b = 255 - *(a + x*stride + RuleIdx)*255/3;
                        //g = 0;
                        *scanLine++ = qRgb(r,r,r);
                    }
                }
            }
            if (cnt % STEP == 0) {
                emit renderedImage(image,rect);
            }
            if (unchanged_cas == 2) {
                CA1D_vec::iterator ca_i, ca_e;
                ca_e = cas.end();
                for (ca_i = cas.begin(); ca_i != ca_e; ca_i++) {
                    CA1D_ptr ca = *ca_i;
                    std::cout << " " << ca->last_change_at;
                }
                break;
            }
        }
        if (restart)
            break;
        if (abort)
            return;
    }
}

CA1D_ptr RenderThread::initCA(void) {

    Nbhs_ptr nbhs(new Nbhs);
    Coords c = Coords();
    for (int i = -1; i != 2; i++) {
        for (int j = -1; j != 2; j++) {
            c.push_back(Coord(i,j));
        }
    }
    nbhs->push_back(c);

    Rules *rules = new Rules;
    Rule *rule = new Rule;
    TSpecificFunctor<Rule> *ruleF = new TSpecificFunctor<Rule> (rule, &Rule::apply);
    rules->push_back(ruleF);

    State_ptr st(new State(Coord(LAT_W, LAT_H), nbhs, *rules));
    st->read_state_from(*state);


    return CA1D_ptr(new CA1D (st));

}


CA1D_ptr RenderThread::initHOCA(void) {

    Nbhs_ptr nbhs(new Nbhs);
    Coords c = Coords();
    for (int i = -1; i != 2; i++) {
        for (int j = -1; j != 2; j++) {
            c.push_back(Coord(i,j));
        }
    }
    nbhs->push_back(c);


    Rules *rules = new Rules;
    for (int i = 0; i != 100; i++) {
        HOCARule *rule = new HOCARule;
        TSpecificFunctor<HOCARule> *ruleF = new TSpecificFunctor<HOCARule> (rule, &HOCARule::apply);
        rules->push_back(ruleF);
    }
    for (int i = 0; i != 5; i++) {
        ZRule *zrule = new ZRule;
        TSpecificFunctor<ZRule> *zruleF = new TSpecificFunctor<ZRule> (zrule, &ZRule::apply);
        rules->push_back(zruleF);
    }

    State_ptr st(new State(Coord(LAT_W, LAT_H), nbhs, *rules));
    //st->clear();
    st->read_state_from(*state);

    return CA1D_ptr(new CA1D (st));

}
