
#include "rules.h"
#include <string.h>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <cstring>


State::State(const Coord &dims, const Nbhs_ptr n, const Rules r, const unsigned int params_):
        width(dims.first), height(dims.second), params(params_), x_stride(params_), y_stride(x_stride*dims.first),
        arr_size(y_stride*dims.second), nbhs(n), rules(r), arr(new unsigned char [arr_size]) {
}

void State::apply(State &next) {
    TFunctor *rule;
    next.changes = 0;
    next.live_cells = 0;
    for (unsigned int y = 0; y < height; y++) {
        for (unsigned int x = 0; x < width; x++) {
            rule = rules.at(arr[x*x_stride + y*y_stride + RuleIdx]);
            rule->apply(*this, next, x, y);
        }
    }
}

unsigned int State::outer_sum(unsigned int x, unsigned int y, unsigned char &state, unsigned long int &arr_index) {
    arr_index = x*x_stride + y*y_stride;
    state = arr[arr_index];

    int nbh_i = arr[arr_index + NbhIdx];   // neighbourhood index is the first parameter
    Coords::iterator ni, ne;          // [n]eighbourhood [i]terator and [n]eighbourhood [e]nd.
    ne = (*nbhs)[nbh_i].end();
    long int comp_x = 0, comp_y = 0;
    unsigned int sum = 0;
    for (ni = (*nbhs)[nbh_i].begin(); ni != ne; ni++) {
        if (ni->first == 0 && ni->second == 0) {
            continue;
        }
        comp_x = ((x + ni->first)*x_stride + y_stride) % y_stride;
        comp_y = ((y + ni->second)*y_stride + arr_size) % arr_size;
        sum += arr[comp_x + comp_y];
    }
    return sum;
}

unsigned char & State::mod_idx(unsigned int x, unsigned int y, unsigned int mod, Parameter p) {
    unsigned long int arr_index = (x - x%mod)*x_stride + (y - y%mod)*y_stride + p;
    return arr[arr_index];
}

void State::print(void) {
    for (unsigned int y = 0; y < height; y++) {
        for (unsigned int x = 0; x < width; x++) {
            std::cout << std::setw(4) << static_cast<int>(arr[x*x_stride + y*y_stride]) << " ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}

State * State::clone(void) {
    Coord c = Coord(width, height);
    State * st = new State(c, nbhs, rules, params);
    std::memcpy(st->get(), arr.get(), arr_size);
    return st;
}

void State::random(unsigned int perc, unsigned int param) {
    unsigned char * b = this->begin() + param;
    unsigned char * e = this->end();
    if (perc == 0) {
        while (b < e) {
            *b = 0;
            b += x_stride;
        }
        return;
    }
    while (b < e) {
        if (rand() % 100 >= perc) {
            *b = 0;
        } else {
            *b = 1;
        }
        b += x_stride;
    }
}

void State::clear(int param) {
    if (param != -1) {
        random(0, param);
    } else {
        for (int i = 0; i != params; i++) {
            random(0, i);
        }
    }
}

void State::read_state_from(State &st) {
    unsigned char * b = this->begin();
    unsigned char * e = this->end();
    unsigned char * ob = st.begin();
    unsigned char * oe = st.end();
    while (b != e) {
        *b++ = *ob++;
    }
}

unsigned char Rule::apply(State &src, State &tar, unsigned int x, unsigned int y) {
    unsigned char state = 0;
    unsigned long int arr_index = 0;
    unsigned int outer_sum = src.outer_sum(x, y, state, arr_index);  // sets state to state of central cell
    unsigned char new_state = 0;
    unsigned char old_state = src.get()[arr_index + StateIdx];
    if (outer_sum == 2) {
        new_state = old_state;
    } else if (outer_sum == 3) {
        new_state = 1;
    }
    tar.get()[arr_index + StateIdx] = new_state;
    if (new_state != old_state) {
        tar.changes++;
    }
    if (new_state == 1) {
        tar.live_cells++;
    }

    return new_state;
}

unsigned char HOCARule::apply(State &src, State &tar, unsigned int x, unsigned int y) {
    unsigned char state = 0;
    unsigned long int arr_index = 0;
    unsigned int outer_sum = src.outer_sum(x, y, state, arr_index);  // sets state to state of central cell
    unsigned char new_state = 0;
    unsigned char old_state = src.get()[arr_index + StateIdx];
    unsigned char rule_idx = src.get()[arr_index + RuleIdx];
    if (outer_sum == 2) {
        new_state = old_state;
    } else if (outer_sum == 3) {
        new_state = 1;
    }
    tar.get()[arr_index + StateIdx] = new_state;

    if (old_state == new_state && new_state == 1) {
        tar.get()[arr_index + RuleIdx] = src.get()[arr_index + RuleIdx] + 1;
    } else {
        tar.get()[arr_index + RuleIdx] = 0;
    }
    if (new_state != old_state) {
        tar.changes++;
    }
    if (new_state == 1) {
        tar.live_cells++;
    }

    return new_state;
}

unsigned char ZRule::apply(State &src, State &tar, unsigned int x, unsigned int y) {
    unsigned long int arr_index = x*src.x_stride + y*src.y_stride;
    unsigned char old_state = src.get()[arr_index + StateIdx];
    tar.get()[arr_index + StateIdx] = 0;

    unsigned int new_rule_idx = src.get()[arr_index + RuleIdx] + 1;
    if (tar.number_of_rules() == new_rule_idx) {
        tar.get()[arr_index + RuleIdx] = 0;
    } else {
        tar.get()[arr_index + RuleIdx] = new_rule_idx;
    }
    if (0 != old_state) {
        tar.changes++;
    }

    return 0;
}


CA1D::CA1D(State_ptr &s): state(State_ptr(s->clone())), next(State_ptr(s->clone())), width(s->width), height(s->height) {

}

State_ptr CA1D::get() {
    state->apply(*next);
    state.swap(next);
    return state;
}



