 
#ifndef RULES_H
#define RULES_H

#include <vector>
#include <boost/shared_array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/ptr_container/ptr_vector.hpp>


enum Parameter { StateIdx, NbhIdx, RuleIdx, ChangesIdx };

class State;
typedef boost::shared_ptr<State> State_ptr;

typedef std::pair<int, int> Coord;
typedef std::vector<Coord> Coords;
typedef std::vector<Coords> Nbhs;

typedef boost::shared_ptr<Nbhs> Nbhs_ptr;


typedef unsigned char (*fptr)(State &src, State &tar, unsigned int x, unsigned int y);

class TFunctor
{
public:
  virtual unsigned char apply(State &src, State &tar, unsigned int x, unsigned int y)=0;        // call using function
};

// derived template class
template <class TClass> class TSpecificFunctor : public TFunctor
{
private:
  unsigned char (TClass::*fpt)(State &src, State &tar, unsigned int x, unsigned int y);   // pointer to member function
  TClass* pt2Object;                  // pointer to object

public:
  // constructor - takes pointer to an object and pointer to a member and stores
  // them in two private variables
  TSpecificFunctor(TClass* _pt2Object, unsigned char(TClass::*_fpt)(State &src, State &tar, unsigned int x, unsigned int y))
     { pt2Object = _pt2Object;  fpt=_fpt; }

  // override function "Call"
  virtual unsigned char apply(State &src, State &tar, unsigned int x, unsigned int y)
    { return (*pt2Object.*fpt)(src, tar, x, y);}             // execute member function
};

typedef std::vector < TFunctor* > Rules;

class Rule;

class State
{
public:
    State(const Coord &dims, const Nbhs_ptr n, const Rules r, const unsigned int params_=5);

    const unsigned int width, height, params, x_stride, y_stride, arr_size;

    unsigned char * begin(void)   { return arr.get(); }
    unsigned char * end(void)     { return arr.get() + arr_size; }
    unsigned char * get(void)     { return arr.get(); }
    unsigned char & mod_idx(unsigned int x, unsigned int y, unsigned int mod, Parameter p);

    void print(void);
    State * clone(void);
    unsigned char * scanLine(unsigned int y) { return arr.get() + y*y_stride; }

    void apply(State &next);
    unsigned int outer_sum(unsigned int x, unsigned int y, unsigned char &state, unsigned long int &arr_index);
    void random(unsigned int perc=0, unsigned int param=0);
    void clear(int param=-1);
    void read_state_from(State &st);
    unsigned int number_of_rules(void) { return rules.size(); }
    unsigned int changes, live_cells;

  private:
    const Nbhs_ptr nbhs;
    const Rules rules;
    boost::shared_array<unsigned char> arr;
};

class Rule
{
 public:
    Rule() {}

    unsigned char apply(State &src, State &tar, unsigned int x, unsigned int y);
};


class HOCARule
{
 public:
    HOCARule() {}

    unsigned char apply(State &src, State &tar, unsigned int x, unsigned int y);
};

class ZRule
{
 public:
    ZRule() {}

    unsigned char apply(State &src, State &tar, unsigned int x, unsigned int y);
};

class CA1D
{
 public:
    CA1D(State_ptr &s);
    State_ptr get(void);
    unsigned int last_changes, unchanged_for, last_change_at;
    
 private:
    boost::shared_ptr<State> state, next;

 public:
   const unsigned int width, height;
};

typedef boost::shared_ptr<CA1D> CA1D_ptr;

#endif
