#!/usr/bin/env python


import numpy as np
import math
import sys

def radiusNbh(r=1):
    return range(-r, r+1)


def connect(l_below, l_above, M):
    l_below.up = M
    l_below.up.above = l_above
    l_above.down = M
    l_above.down.below = l_below


def statify(state_list, max_state):
    ret = 0
    shift = math.log(max_state+1)/math.log(2)
    if int(shift) != shift:
        shift = int(shift) + 1
    else:
        shift = int(shift)
    for n in state_list:
        ret = (ret << shift) + n
    return ret



class null_map(object):
    def __init__(self, **kwargs):
        self.target = kwargs.get("target", None)

    def set_target(self, target):
        self.target = target

    def up(self, x):
        # Prototype function, should be redefined by descendants
        return ()

    def down(self, x):
        # Prototype function, should be redefined by descendants
        return ()


class Lattice:
    def __init__(self, twoNbh, twoRule, transition_map,
                 max_state=1, width=300, dtypes=("uint8",)):
        self.nbh = np.zeros((width,), dtype="uint8")
        self.twoNbh = twoNbh
        self.rule = np.zeros((width,), dtype="uint8")
        self.twoRule = twoRule
        self.tmap = transition_map
        self.max_state = max_state
        self.width = width
        self.dtypes = dtypes[:]
        self.state = []
        self.max = []
        for dtype in dtypes:
            self.state.append(np.zeros((width,), dtype=dtype))
            self.max.append(0)
        self.up = null_map()
        self.down = null_map()

    def __getitem__(self, item):
        return [x[item%self.width] for x in self.state]

    def __setitem__(self, item, value):
        for i in range(len(self.state)):
            self.state[i][item%self.width] = value[i]

    def contains(self, x):
        return self.down.down(x)

    def down_states(self, x, num=0):
        return [self.down.below.state[num][x] for x in self.contains(x)]

    def organs(self, x):
        return self.up.up(x)

    def up_states(self, x, num=0):
        if type(num) in (type([]), type(())):
            return [[self.up.above.state[y][x] for y in num] \
                        for x in self.organs(x)]
        return [self.up.above.state[num][x] for x in self.organs(x)]

    def next(self):
        ret = Lattice(self.twoNbh, self.twoRule, self.tmap, 
                      self.max_state, self.width, self.dtypes)
        for x in range(self.width):
            rule = self.twoRule[self.rule[x]]
            ret[x] = rule(self, x)
            (ret.nbh[x], ret.rule[x]) = self.tmap(self, x)
        for i in range(len(ret.state)):
            ret.max[i] = max(np.max(ret.state[i]), self.max[i])
        ret.up = self.up
        ret.down = self.down
        return ret

    def states(self, x):
        nbh = self.twoNbh[self.nbh[x]]
        return ([self[x+n] for n in nbh], self.nbh[x], self.rule[x])

    def nbh_state(self, x, layer=0):
        nbh = self.twoNbh[self.nbh[x]]
        return statify([self.state[layer][(x+n)%self.width] for n in nbh], 
                       self.max_state)

    def sum(self, x, num=0):
        nbh = self.twoNbh[self.nbh[x]]
        layer = self.state[num]
        return sum([layer[(x+n)%self.width] for n in nbh])

    def outer_sum(self, x):
        return self.sum(x) - self[x]

    def random(self, nums=(0,), **kwargs): 
        if kwargs.get("nbh", True):
           self.nbh = np.random.randint(0, len(self.twoNbh), (self.width,))
        if kwargs.get("rule", True):
            self.rule = np.random.randint(0, len(self.twoRule), (self.width,))
        for n in nums:
            self.state[n] = np.random.randint(0, self.max_state+1, (self.width,))

    def copy(self):
        ret = Lattice(self.twoNbh, self.twoRule, self.tmap, 
                      self.max_state, self.width, self.dtypes,
                      self.identifier)
        ret.up = self.up
        ret.down = self.down
        ret.rule = self.rule.copy()
        ret.nbh = self.nbh.copy()
        for i in range(len(self.state)):
            ret.state[i] = self.state[i].copy()
        return ret


class LatticeStack:
    def __init__(self, lattices, identifier=""):
        if not type(lattices) in (type(()), type([])):
            lattices = (lattices,)
        self.lattices = lattices
        self.identifier = identifier

    def __getitem__(self, t):
        return self.lattices[t]

    def next(self):
        new_lattices = []
        for l in self.lattices:
            new_lattices.append(l.next())
            if len(new_lattices) > 1:
                connect(new_lattices[-2], new_lattices[-1], l.down)
        return LatticeStack(new_lattices, self.identifier)

    def random(self, nums_arr=None, **kwargs):
        for i in range(len(self.lattices)):
            nums = (0,)
            if nums_arr != None and len(nums_arr) > i:
                nums = nums_arr[i]
            self.lattices[i].random(nums, **kwargs)

    def copy(self):
        ret = []
        for l in self.lattices:
            ret.append(l.copy())
            if len(ret) > 1:
                connect(ret[-2], ret[-1], l.down)
        return LatticeStack(ret, self.identifier)
        

       
class Pattern:
    def __init__(self, n=100, t=0):
        self.t = t
        self.n = n
        self.pat = []

    def __getitem__(self, t):
        return self.pat[t]

    def fill(self, state):
        self.pat = []
        while len(self.pat) < self.n:
            self.pat.append(self.pat[-1].next())

    def len(self):
        return len(self.pat)


class CA:
    def __init__(self, state, n=100, t=0):
        self.state = state
        self.pattern = Pattern(n, t)

    def get(self):
        self.pattern.fill(self.state)
        self.state = self.pattern[-1]
        return self.pattern

    def set_state(self, state):
        self.state = state
        

class NullTMap:
    def __call__(self, lattice, x):
        return (0,0)

class WolframRule:
    def __init__(self, rule):
        if type(rule) == type(""):
            self.rule = int(rule, 2)
        else:
            self.rule = rule

    def __call__(self, lattice, x):
        st = lattice.nbh_state(x)
        return ((self.rule >> st) & 1,)

class WolframMapRule:
    def __init__(self, map):
        self.map = map

    def __call__(self, lattice, x):
        s = lattice.nbh_state(x)
        #print s
        return (self.map[s],)
    
                                         
class SumMapRule:
    def __init__(self, map):
        if type(map) == type(""):
            map = tuple([int(x) for x in map])
        if type(map) == type(0):
            (map, i) = ([], map)            
            while i > 0:
                map.append(i & 1)
                i = i >> 1
        self.map = map

    def __call__(self, lattice, x):
        sum = lattice.sum(x)
        if sum > len(self.map) - 1:
            return (0,)
        return (self.map[sum],)


class SumWMapRule:
    def __init__(self, map, weights):
        self.map = map
        self.weights = map

    def __call__(self, lattice, x):
        s = sum([self.weights[i[0]] for i in lattice.states(x)[0]])
        if s > len(self.map) - 1:
            return (0,)
        return (self.map[abs(s)],)


class TotExTMap:
    def __init__(self, ca, cb):
        self.ca = ca
        self.cb = cb

    def __call__(self, lattice, x):
        s = lattice.sum(x)
        r = lattice.rule[x]
        if r == 1 and s == self.cb:
            return (0, 0)
        if r == 0 and s == self.ca:
            return (0, 1)
        return (0, r)


class SumMapTMap:
    def __init__(self, map):
        if type(map) == type(""):
            map = tuple([int(x) for x in map])
        if type(map) == type(0):
            (map, i) = ([], map)            
            while i > 0:
                map.append(i & 1)
                i = i >> 1
        self.map = map

    def __call__(self, lattice, x):
        sum = lattice.sum(x)
        if sum > len(self.map) - 1:
            return (0, 0)
        return (0, self.map[sum])


class EquiTotTMap:
    def __init__(self, smin, smax, rmin, rmax):
        self.smin = smin
        self.smax = smax
        self.nmin = 2*rmin+1
        self.nmax = 2*rmax+1
        
    def __call__(self, lattice, x):
        n = lattice.nbh[x]
        m = len(lattice.twoNbh[n])
        s = lattice.sum(x)
        if s < self.smin and m < self.nmax:
            return (n+1, 0)
        if s > self.smax and m > self.nmin:
            return (n-1, 0)
        return (n, 0)


class ModMap(null_map):
    def __init__(self, mod, **kwargs):
        super(ModMap, self).__init__(**kwargs)
        self.mod = mod

    def up(self, x):
        return (x/self.mod,)

    def down(self, x):
        return range(x*self.mod, (x+1)*self.mod)


def enlarge(data, factor):
    ret = np.zeros([factor*x for x in data.shape], data.dtype)
    np.transpose(ret.reshape(data.shape[0], factor))[:] = data
    return ret

if __name__ == "__main__":
    two_nbh = ( radiusNbh(2), )
    a = SumMapRule((0,0,1,1,0,0))
    b = SumMapRule((0,1,0,0,0,0))
    two_rule = (a, b)
    tmap = TotExTMap(3, 3)
    l = Lattice(two_nbh, two_rule, tmap)
    l.random()
    ls = LatticeStack((l,))
    ca = CA(ls)
    ca.get()
