/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at http://www.qtsoftware.com/contact.
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QtGui>
#include <iostream>
#include <math.h>

#include "displaywidget.h"


const double DefaultX = 0;
const double DefaultY = 0;
const double DefaultScale = 1;
//const double DefaultX = -200;
//const double DefaultY = -150;
//const double DefaultScale = 0.5;

const double ZoomInFactor = 0.8f;
const double ZoomOutFactor = 1 / ZoomInFactor;
const int ScrollStep = 20;

DisplayWidget::DisplayWidget(QWidget *parent)
    : QWidget(parent)
{
    screen_x = DefaultX;
    screen_y = DefaultY;
    pixmapScale = 1;
    rules.push_back(30); rules.push_back(33);
    curScale = DefaultScale;

    qRegisterMetaType<QImage>("QImage");
    qRegisterMetaType<QRect>("QRect");
    connect(&thread, SIGNAL(renderedImage(const QImage &, const QRect &)),
            this, SLOT(updatePixmap(const QImage &, const QRect &)));
    setWindowTitle(tr("DisplayWidget"));
#ifndef QT_NO_CURSOR
    setCursor(Qt::CrossCursor);
#endif
    resize(1000, 500);
}

void DisplayWidget::rule1_changed(int n) {
}

void DisplayWidget::paintEvent(QPaintEvent * /* event */)
{
    QPainter painter(this);
    painter.fillRect(rect(), Qt::black);

    if (pixmap.isNull()) {
        painter.setPen(Qt::white);
        painter.drawText(rect(), Qt::AlignCenter,
                         tr("Rendering initial image, please wait..."));
        return;
    }

    if (curScale == pixmapScale) {
        painter.drawPixmap(pixmapOffset, pixmap);
    } else {
        double scaleFactor = pixmapScale / curScale;
        int newWidth = int(pixmap.width() * scaleFactor);
        int newHeight = int(pixmap.height() * scaleFactor);
        int newX = pixmapOffset.x() + (pixmap.width() - newWidth) / 2;
        int newY = pixmapOffset.y() + (pixmap.height() - newHeight) / 2;

        painter.save();
        painter.translate(newX, newY);
        painter.scale(scaleFactor, scaleFactor);
        QRectF exposed = painter.matrix().inverted().mapRect(rect()).adjusted(-1, -1, 1, 1);
        painter.drawPixmap(exposed, pixmap, exposed);
        painter.restore();
    }

//    QString text = tr("Use mouse wheel or the '+' and '-' keys to zoom. "
//                      "Press and hold left mouse button to scroll.");
//    QFontMetrics metrics = painter.fontMetrics();
//    int textWidth = metrics.width(text);
//
//    painter.setPen(Qt::NoPen);
//    painter.setBrush(QColor(0, 0, 0, 127));
//    painter.drawRect((width() - textWidth) / 2 - 5, 0, textWidth + 10,
//                     metrics.lineSpacing() + 5);
//    painter.setPen(Qt::white);
//    painter.drawText((width() - textWidth) / 2,
//                     metrics.leading() + metrics.ascent(), text);
}



void DisplayWidget::resizeEvent(QResizeEvent * /* event */)
{
    pixmap = QPixmap(size());
    pixmap.fill(QColor(255,255,255).rgb());
    thread.render(screen_x, screen_y, size());
}

void DisplayWidget::keyPressEvent(QKeyEvent *event)
{
    switch (event->key()) {
    case Qt::Key_Plus:
        zoom(ZoomInFactor);
        break;
    case Qt::Key_Minus:
        zoom(ZoomOutFactor);
        break;
    case Qt::Key_Left:
        scroll(-ScrollStep, 0);
        break;
    case Qt::Key_Right:
        scroll(+ScrollStep, 0);
        break;
    case Qt::Key_Down:
        scroll(0, -ScrollStep);
        break;
    case Qt::Key_Up:
        scroll(0, +ScrollStep);
        break;
    default:
        QWidget::keyPressEvent(event);
    }
}

void DisplayWidget::wheelEvent(QWheelEvent *event)
{
    int numDegrees = event->delta() / 8;
    double numSteps = numDegrees / 15.0f;
    zoom(pow(ZoomInFactor, numSteps));
}

void DisplayWidget::mousePressEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton)
        lastDragPos = event->pos();
}

void DisplayWidget::mouseMoveEvent(QMouseEvent *event)
{
    if (event->buttons() & Qt::LeftButton) {
        pixmapOffset += event->pos() - lastDragPos;
        lastDragPos = event->pos();
        update();
    }
}

void DisplayWidget::mouseReleaseEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton) {
        pixmapOffset += event->pos() - lastDragPos;
        lastDragPos = QPoint();

        int deltaX = (width() - pixmap.width()) / 2 - pixmapOffset.x();
        int deltaY = (height() - pixmap.height()) / 2 - pixmapOffset.y();
        scroll(deltaX, deltaY);
    }
}

void DisplayWidget::updatePixmap(const QImage &image, const QRect &rect)
{
    //printf("updating...\n"); fflush(stdout);
    if (!lastDragPos.isNull())
        return;
    QRect src = rect.translated(-rect.x(), -rect.y());
    QPoint ul = QPoint(-screen_x, -screen_y);
    QRect tar = rect.translated(ul);

    QPainter painter(&pixmap);
    painter.drawImage(tar, image, src);
//    painter.setPen(QPen(QBrush(QColor(0,255,255,100)),1));
//    painter.setBrush(QBrush(Qt::NoBrush));
//    painter.drawRect(tar);
    pixmapOffset = QPoint();
    lastDragPos = QPoint();
    painter.end();
    update();
}

void DisplayWidget::zoom(double zoomFactor)
{
    curScale *= zoomFactor;
    update();
    thread.render(screen_x, screen_y, size());
}

void DisplayWidget::scroll(int deltaX, int deltaY)
{
    screen_x += deltaX * curScale;
    screen_y += deltaY * curScale;
    update();
    thread.render(screen_x, screen_y, size());
}

void DisplayWidget::closeEvent(QCloseEvent *)
{
    ;
}

