/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at http://www.qtsoftware.com/contact.
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QtGui>
#include <iostream>
#include <math.h>

#include "renderthread.h"

#include "rules.h"

// The dimensions of each lattice
const unsigned int LAT_W = 80;
const unsigned int LAT_H = 60;

// Print column and row text labels?
const bool ANNOTATE = true;

// The room left to the left and above for the row and
// column text labels, if ANNOTATE == true
unsigned int X_MARGIN = 40;
unsigned int Y_MARGIN = 15;

// Which values should the first parameter vary through?
const double X_VALS [] = {0.5, 1, 1.5, 2, 3, 5, 8, 12};
// Which values should the second parameter vary through?
const double Y_VALS [] = {0.5, 1, 2, 4, 6};
// Which values should the third parameter vary through?
const unsigned int P_VALS [] = {20,22,25,32,48};   //9,11,14,16,19

// Update the picture on screen each STEPth time step
const unsigned int STEP = 1;
// Store an image of the pattern each STORE_STEPth time step
const unsigned int STORE_STEP = 1200;
// Run for RUN_STEPS time step before starting over with a new vaule from P_VALS
const unsigned int RUN_STEPS = 1201;

CA1D_ptr RenderThread::get_CA(unsigned int x, unsigned int y, unsigned int p) {
    // Change this function to alter what parameters get varied with the values
    // in X_VALS, Y_VALS and P_VALS. The order of the parameters given to
    // initCA is: k1, k2, g, n
    return initCA(X_VALS[x], Y_VALS[y], P_VALS[p], 99);
}


RenderThread::RenderThread(QObject *parent)
    : QThread(parent), state(State_ptr(new State(Coord(LAT_W,LAT_H), Nbhs_ptr(), Rules())))
{
    restart = false;
    abort = false;
    main_working = false;
    state->clear();
    state->random(100, StateIdx, 0, 100);
    working_mutex.unlock();
}


RenderThread::~RenderThread()
{
    mutex.lock();
    abort = true;
    condition.wakeOne();
    mutex.unlock();

    wait();
}

void RenderThread::render(int x, int y, QSize resultSize)
{
    QMutexLocker locker(&mutex);

    this->x = x;
    this->y = y;
    this->resultSize = resultSize;

    if (!isRunning()) {
        start(LowPriority);
    } else {
        restart = true;
        condition.wakeOne();
    }
}

void RenderThread::set_working(bool w) {
    QMutexLocker locker(&working_mutex);
    main_working = w;
    if (main_working == false)
        working.wakeOne();
}

void RenderThread::run()
{
    const unsigned int X_NUM = sizeof(X_VALS)/sizeof(X_VALS[0]);
    const unsigned int Y_NUM = sizeof(Y_VALS)/sizeof(Y_VALS[0]);
    if (!ANNOTATE) {
        X_MARGIN = 0;
        Y_MARGIN = 0;
    }
    std::cout << X_MARGIN << std::endl;
    typedef std::vector<CA1D_ptr> CA1D_vec;

    for (int g = 0; g != sizeof(P_VALS)/sizeof(P_VALS[0]); g ++) {
        std::cout << std::endl << "p_val: " << P_VALS[g] << std::endl << "i: ";

        CA1D_vec ca_vec;
        for (int i = 0; i < X_NUM; i++) {
            for (int j = 0; j < Y_NUM; j++) {
                // Here we may change what parameters to vary
                CA1D_ptr p = get_CA(i, j, g);
                p->x = i; p->y = j;
                ca_vec.push_back(p);
            }
        }

        unsigned int ca_width = ca_vec[0]->width;
        unsigned int ca_height = ca_vec[0]->height;
        QRect ca_rect(0,0,ca_width,ca_height);

        unsigned int width = (ca_width+2)*X_NUM + X_MARGIN;
        unsigned int height = (ca_height+2)*Y_NUM + Y_MARGIN;
        QRect large_rect(0,0,width,height);

        QImage bm(large_rect.size(), QImage::Format_RGB32);
        QImage image(ca_rect.size(), QImage::Format_RGB32);
        State_ptr st;


        int j = 1;
        if (restart)
            break;
        while (j*STEP < RUN_STEPS) {
            bm.fill(QColor(255,255,255).rgb());

            CA1D_vec::iterator ca_i, ca_e;
            ca_e = ca_vec.end();
            for (ca_i = ca_vec.begin(); ca_i != ca_e; ca_i++) {
                if (abort)
                    return;
                CA1D_ptr ca = *ca_i;
                if ((j*STEP) % 500 == 0) {
                    ca->get(false)->clear(ChangesIdx);
                }
                for (int i = 0; i != STEP; i++) {
                    st = ca->get();
                }


                const unsigned int *a;
                unsigned int r;
                unsigned int max = ca->n;
                unsigned int stride = st->x_stride;
                for (int y = 0; y < ca_height; ++y) {
                    uint *scanLine =
                            reinterpret_cast<uint *>(image.scanLine(y));
                    a = st->scanLine(y);
                    for (int x = 0; x != ca_width; x++) {
                        //*scanLine++ = qRgb(252,123,0);
                        r = *(a + x*stride);
                        //std::cout << "r: " << r;
                        if (r > 0) {
                            if (r < max) {
                                r = 200 - r*100/max;
                            } else {
                                r = 0;
                            }
                        } else {
                            r = 255;
                        }
                        //std::cout << "->" << r << std::endl;
                        *scanLine++ = qRgb(r,r,r);
                    }
                }
                QPainter *p = new QPainter();
                p->begin(&bm);
                QPoint ul = QPoint(ca->x*(ca_width+2)+X_MARGIN, ca->y*(ca_height+2)+Y_MARGIN);
                QRect tar = ca_rect.translated(ul);
                p->drawImage(tar, image, ca_rect);
                if (ANNOTATE) {
                    if (ca->x == 0 || ca->y == 0) {
                        QFont font(p->font());
                        font.setPixelSize(10);
                        p->setFont(font);
                        QFontMetrics metrics = p->fontMetrics();
                        p->setPen(Qt::black);
                        QString text;
                        if (ca->x == 0) {
                            text = QString("k2=") + QString().setNum(ca->k2);
                            p->drawText(ul+QPoint(-ul.x(),ca_height/2), text);
                        }
                        if (ca->y == 0) {
                            text = QString("k1=") + QString().setNum(ca->k1);
                            p->drawText(ul+QPoint(0,metrics.leading()+metrics.ascent())+QPoint(ca_width/2-metrics.width(text),-ul.y()), text);
                        }
                    }
                }
                p->end();
                delete p;
            }

            emit renderedImage(bm, large_rect);
            std::cout << j*STEP << " " << std::flush;
            //sleep(10);
            if ((j*STEP) % STORE_STEP == 0) {
                bm.save(QString("hodgepodge-") + QString().setNum(P_VALS[g]) + QString("-")  + QString().setNum(j*STEP) + QString(".jpg"), "JPEG");
            }
            j++;
        }

    }
}

Coord rotateCoord(Coord c, double deg, double rad=1.9) {
    double a = cos(deg);
    double b = sin(deg);
    double x = rad*(a*c.first - b*c.second);
    double y = rad*(b*c.first + a*c.second);
    return Coord(static_cast<int>(x), static_cast<int>(y));
}

CA1D_ptr RenderThread::initCA(double k1, double k2, int g, int n) {
    Nbhs_ptr nbhs(new Nbhs);
    Coords c = Coords();
    c.push_back(Coord(0,0));
    for (double a=0; a < 6.28; a += 3.14/4) {
        Coord coord = rotateCoord(Coord(1,0), a);
        c.push_back(coord);
        //std::cout << "(" << coord.first << ", " << coord.second << ") ";
    }
    nbhs->push_back(c);

    Rules *rules = new Rules;
    Rule *rule = new Rule(k1, k2, g, n);
    //std::cout << "k1: " << k1 << " k2: " << k2 << std::endl;
    TSpecificFunctor<Rule> *ruleF = new TSpecificFunctor<Rule> (rule, &Rule::apply);
    rules->push_back(ruleF);
//    Rule *zrule = new ZRule(k1, k2, g, n);
//    TSpecificFunctor<ZRule> *zruleF = new TSpecificFunctor<ZRule> (zrule, &ZRule::apply);
//    rules->push_back(zruleF);

    State_ptr st(new State(Coord(LAT_W,LAT_H), nbhs, *rules));
    st->read_state_from(*state);
    //st->random(100,StateIdx,0, 101);
    //std::cout << st->get()[0] << std::endl;

    CA1D_ptr p = CA1D_ptr(new CA1D (st));
    p->k1 = k1;
    p->k2 = k2;
    p->g = g;
    p->n = n;

    return p;
}
