#!/usr/bin/env python

CA1D_DEF_SIZE = 35
CA1D_DEF_RULE = 30
CA1D_DEF_INITIAL = "*1*"
CA1D_DEF_EDGE = "01110010"
CA1D_DEF_EDGE_START = 8
CA_ITERATIONS = 16

import numpy as np
from random import randint

def edge_wrap(L, key):
    return L[key]

def edge_alive(L, key):
    return 1

def edge_dead(L, key):
    return 0

frame_alive = edge_alive

def frame_none(L, key):
    return L.N[key]

def frame_pattern(L, key):
    start = CA1D_DEF_EDGE_START
    i = L.iteration
    if i < start:
        return 1
    if i >= start + len(CA1D_DEF_EDGE):
        return 0
    return int(CA1D_DEF_EDGE[i-start])

class CA1D_Rule(object):
    def __init__(self, rule, **kwargs):
        self.rule = rule
        self.edgefunc = kwargs.get("edgefunc", edge_wrap)
        self.framefunc = kwargs.get("framefunc", frame_none)
        self.probs = kwargs.get("probs", frame_none)
        self.shift = kwargs.get("shift", 0)

    def __call__(self, L):
        size = L.size
        rightedge = size - 1
        if L.iteration == 0:
            L.N = L.L.copy()
            L[0] = self.framefunc(L, 0)
            L[rightedge] = self.framefunc(L, rightedge)
            return
        i = 0
        ef = self.edgefunc
        state = (L[self.shift-1] << 1) + L[self.shift]
        while i < size:
            nextcell = L[self.shift+i+1]
            state = ((state << 1) & 7) + nextcell
            L[i] = (self.rule >> state) & 1
            if i == 0 or i == rightedge:
                L[i] = self.framefunc(L, i)
            L[i] = self.probs(L, i)
            i += 1

class CA1D_Lattice(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.size = kwargs.get("size", CA1D_DEF_SIZE)
        self.initial = (kwargs.get("initial", CA1D_DEF_INITIAL)).split("*")
        self.invert = kwargs.get("invert", False)
        self.random = kwargs.get("random", False)
        self.iteration = 0
        self.init_lattice()

    def init_lattice(self):
        self.L = np.zeros(dtype=int, shape=(self.size,))
        self.N = np.zeros(dtype=int, shape=(self.size,))
        if self.random:
            for i in range(self.size):
                self.L[i] = randint(0,1)
        else:
            for (i,s) in enumerate(self.initial[0]):
                self.L[i] = int(s)
            for (i,s) in enumerate(self.initial[1]):
                self.L[self.size/2-len(self.initial[1])/2+i] = int(s)
            for (i,s) in enumerate(self.initial[2]):
                self.L[len(self.L)-len(self.initial[2])+i] = int(s)
        if self.invert:
            self.L = 1 - self.L

    def apply_rule(self, rule):
        rule(self)
        self.flip()
        return self.L

    def flip(self):
        (self.L, self.N) = (self.N, self.L)
        self.iteration += 1
        
    def __getitem__(self, key):
        return self.L[key % self.size]

    def __setitem__(self, key, value):
        self.N[key] = value

    def copy(self):
        l = self.__class__(**self.kwargs)
        l.L = self.L.copy()
        l.N = self.N.copy()
        l.iteration = self.iteration
        return l


class CA(object):
    def __init__(self, **kwargs):
        self.rule = kwargs.get("rule", None)
        self.lattice = kwargs.get("lattice", None)
        self.max_iter = kwargs.get("iterations", CA_ITERATIONS)
        self.skip = kwargs.get("skip", 0)

class CA1D(CA):
    def __init__(self, **kwargs):
        super(CA1D, self).__init__(**kwargs)
        if self.rule == None:
            self.rule = CA1D_Rule(CA1D_DEF_RULE)
        if self.lattice == None:
            self.lattice = CA1D_Lattice(CA1D_DEF_INITIAL)
        
    def next(self):
        while self.lattice.iteration < self.skip:
            self.lattice.apply_rule(self.rule)
        if self.lattice.iteration == self.skip + self.max_iter:
            raise StopIteration
        return self.lattice.apply_rule(self.rule)

    def __iter__(self):
        return self

def pprint(l):
    r = " |"
    for x in l:
        r += x and "#" or " "
    return r + "|"
        
if __name__ == "__main__":
    l = CA1D_Lattice(size=70, initial="111011*1*110111")
    r = CA1D_Rule(30, framefunc=frame_pattern)
    ca = CA1D(lattice=l, rule=r)
    for l in ca:
        print pprint(l)
