#!/usr/bin/env python

from optparse import OptionParser, OptionValueError
import os
import sys
#from random import randint
#from Queue import Queue
from ca import *

def parse_bin_rule(option, opt, value, parser):
    try:
        rule = int(value, 2)
    except ValueError:
        raise OptionValueError("Argument %s: Not a binary string: %s" %
                               (opt, value))
    parser.values.rule = rule

def parse_options(argv=None):

    if argv == None:
        argv = sys.argv[1:]

    usage = "usage: %prog [options] output.pdf"
    parser = OptionParser(usage=usage)
    parser.add_option("-r", action="store", type="int", dest="rule",
                      help="Rule to use. [%default]", metavar="RULE_NUMBER",
                      default=30)
    parser.add_option("-R", action="callback", type="string", 
                      callback=parse_bin_rule, help="Rule to use, in binary.", 
                      metavar="BINARY_RULE")
    parser.add_option("-i", action="store", type="int", dest="iterations",
                      help="Number of iterations. [%default]", metavar="INT",
                      default=24)
    parser.add_option("-s", action="store", type="int", dest="size",
                      help="Size of lattice. [%default]", default=54,
                      metavar="INT")
    parser.add_option("-c", action="store", type="string", dest="canvas",
                      help="Size of canvas in mm. [%default]", 
                      default="150x200", metavar="WxH")
    parser.add_option("-l", action="store", type="string", dest="label",
                      help="Figure label, defaults to rule number.", 
                      default=None, metavar="QUOTED_STRING")
    parser.add_option("-g", action="store_false", dest="grid",
                      help="Turn off grid", default=True)
    parser.add_option("-F", action="store_true", dest="frame", 
                      default=False, 
                      help="A frame of distinguishing boundary conditions. [%default]")
    parser.add_option("-w", action="store_true", dest="wrap", 
                      default=False, 
                      help="Wrap the lattice. [%default]")
    parser.add_option("-p", action="store", dest="pattern", type="string",
                      default="1", metavar="PATTERN",
                      help="A pattern of initial conditions. [%default]")
    parser.add_option("-I", action="store_true", dest="invert", default=False,
                      help="Invert the starting pattern. [%default]")
    parser.add_option("-f", action="store_true", dest="force", default=False,
                      help="Overwrite existing output file without asking. [%default]")
    
    (options, args) = parser.parse_args(args=argv)
    
    if len(args) != 1:
        parser.print_help()
        print
        sys.exit()
        
    outfilename = args[0]
        
    if os.path.exists(outfilename) and not options.force:
        print "File exits: %s" % outfilename
        sys.exit()
        
    return (options, outfilename)

from reportlab.pdfgen import canvas
from reportlab.lib.units import mm, pica
import numpy as np
pt = pica/12

class Make_PDF:
    def __init__(self, filename, **kwargs):
        pagesize = kwargs.get("pagesize", "150x200")
        grid = kwargs.get("grid", "4x11")
        self.c = canvas.Canvas(filename)
        self.filename = filename
        (self.width, self.height) = [int(x)*mm for x in pagesize.split("x")]
        (self.x_divs, self.y_divs) = [int(x) for x in grid.split("x")]
        (self.x_div, self.y_div) = (-1, 0)
        self.cellspacing = 1.5*mm
        self.x_spacers = (self.x_divs+1)*self.cellspacing
        self.y_spacers = (self.y_divs+1)*self.cellspacing
        self.div_width = (self.width-self.x_spacers)/self.x_divs
        self.div_height = (self.height-self.y_spacers)/self.y_divs
        self.c.setPageSize((self.width, self.height))
        self.square_size = kwargs.get("square_size", 0.8)
        self.font_size = 6
        self.draw_grid = kwargs.get("draw_grid", False)
        self.draw_box = kwargs.get("draw_box", False)
        self.caption = kwargs.get("caption", "")
        self.placement = kwargs.get("placement", "h")
        self.linewidth = kwargs.get("linewidth", 0.1*mm)
        self.texwidth = kwargs.get("width", "")

    def set_options(self, options):
        self.square_size = options.square_size
        self.draw_grid = options.draw_grid
        self.font_size = options.font_size
        self.cellspacing = options.cellspacing
        self.placement = options.placement
        self.caption = options.caption

    def translate(self):
        (x_divs, y_divs) = (self.x_divs, self.y_divs)
        (w, h) = (self.div_width, self.div_height)
        cs = self.cellspacing
        self.x_div += 1
        if self.x_div == x_divs:
            self.x_div = 0
            self.y_div += 1
        if self.y_div == y_divs:
            return False
        self.c.resetTransforms()
        self.c.translate(cs + self.x_div*(w+cs),
                         cs + (y_divs - self.y_div - 1)*(h + cs))
        return True

    def draw_ca(self, ca, bottom_left, size=None, label=""):
        labels = label.split("\n")
        label_space = self.font_size*pt*(len(labels)+0.4)
        if label == "":
            label_space = 0
        c = self.c
        c.saveState()
        c.translate(*bottom_left)
        if size == None:
            (w, h) = (self.div_width, self.div_height)
        else:
            (w, h) = size
        #c.setLineWidth(0.1); c.rect(0,0,w,h,stroke=1,fill=0)
        h -= label_space
        c.translate(0, label_space)
        (x_divs, y_divs) = (ca.lattice.size, ca.max_iter)
        s = min(w/x_divs, h/y_divs)
        c.translate((w - x_divs*s)/2, (h - y_divs*s)/2)
        #c.setLineWidth(0.1); c.rect(0,0,s*x_divs,s*y_divs,stroke=0,fill=1)

        c.setLineWidth(0)
        c.setFillColorRGB(0,0,0)
        m = (1-self.square_size)*0.5*s
        for (y, line) in enumerate(ca):
            for (x, cell) in enumerate(line):
                if cell == 1:
                    c.rect(x*s+m, (y_divs-y-1)*s-m,
                           s-m*2, s-m*2, stroke=0, fill=1)

        if self.draw_grid:
            c.setLineWidth(self.linewidth)
            c.setStrokeColorRGB(0,0,0)
            for i in range(x_divs+1):
                self.c.line(i*s, 0, i*s, y_divs*s)
            for i in range(y_divs+1):
                self.c.line(0, i*s, x_divs*s, i*s)

        if self.draw_box:
            c.setLineWidth(self.linewidth)
            c.setStrokeColorRGB(0,0,0)
            out = 0.5
            drop = 3
            height = (y_divs+2*out)*s
            width = (x_divs+2*out)*s
            self.c.rect(-out*s, -1.5*out*s, width, height)
        c.restoreState()            

        c.setFillColorRGB(0.9,0.9,1)
        c.setLineWidth(0.1); c.rect(0,0,w,label_space,stroke=0,fill=1)
        c.setFillColorRGB(0,0,0)
        c.setFont("Helvetica", self.font_size)
        for (i,l) in enumerate(labels):
            c.drawCentredString(w/2, label_space - (i+1)*self.font_size*pt, l)

    def save(self):
        self.c.showPage()
        self.c.save()
        texfile = os.path.splitext(self.filename)[0] + ".tex"
        f = open(texfile, "w")
        options = ["bb=0mm 0mm %dmm %dmm" % (self.width/mm, self.height/mm)]
        if self.texwidth:
            options.append("width=%s" % self.texwidth)
        d = {"placement": self.placement,
             "noext": "images/" + os.path.splitext(self.filename)[0],
             "stem": os.path.splitext(self.filename)[0],
             "caption": self.caption,
             "options": ",".join(options)}
        f.write(r"""
\begin{figure}[%(placement)s]
  \centering
    \includegraphics[%(options)s]{%(noext)s.pdf}
  \caption{%(caption)s}
  \label{fig:%(stem)s}
\end{figure}
""" % d)
        print "\input{%s}" % d["noext"]


if __name__ == "__main__":
    pass
#    (options, outfilename) = parse_options()
#    pdf = Make_PDF(outfilename)
#    pdf.set_options(options)
#    pdf.draw()
#    pdf.save()
    
