#!/usr/bin/env python


import numpy as np
import math
import random

def vonNeumann():
    return ((0,0), (1,0), (0,1), (-1,0), (0,-1))

def connect(l_below, l_above, M):
    l_below.up = M
    l_below.up.above = l_above
    l_above.down = M
    l_above.down.below = l_below


def statify(state_list, max_state):
    ret = 0
    shift = math.log(max_state)/math.log(2)
    if int(shift) != shift:
        shift = int(shift) + 1
    else:
        shift = int(shift)
    for n in state_list:
        ret = (ret << shift) + n
    return ret

def rot_ccw(tpl):
    return (0*tpl[0]-1*tpl[1],1*tpl[0]+0*tpl[1])

def rot_cw(tpl):
    return (0*tpl[0]+1*tpl[1],-1*tpl[0]+0*tpl[1])

def rot_180(tpl):
    return (-tpl[0], -tpl[1])

def rot_id(tpl):
    return tpl

rot_pri = [rot_id, rot_ccw, rot_cw, rot_180]

class null_map(object):
    def __init__(self, **kwargs):
        self.target = kwargs.get("target", None)
        self.mod = 1

    def set_target(self, target):
        self.target = target

    def up(self, x):
        # Prototype function, should be redefined by descendants
        return ()

    def down(self, x):
        # Prototype function, should be redefined by descendants
        return ()


class Lattice:
    def __init__(self, twoNbh, twoRule, transition_map,
                 **kwargs):
        self.dim = kwargs.get("dim", (120,90))
        self.nbh = np.zeros(self.dim, dtype="uint8")
        self.twoNbh = twoNbh
        self.rule = np.zeros(self.dim, dtype="uint8")
        self.twoRule = twoRule
        self.tmap = transition_map
        self.max_state = kwargs.get("max_state", 1)
        self.dtypes = kwargs.get("dtypes", ("uint8",))[:]
        self.state = []
        for dtype in self.dtypes:
            self.state.append(np.zeros(self.dim, dtype=dtype))
        self.up = null_map()
        self.down = null_map()
        self.inplace = kwargs.get("inplace", False)
        self.on_next = kwargs.get("on_next", "")
        self.kwargs = kwargs
    
    def get_coord(self, coord):
        return (coord[0]%self.dim[0], coord[1]%self.dim[1])

    def swap(self, a, b):
        for i in range(len(self.dtypes)):
            (self.state[i][a[0],a[1]], self.state[i][b[0],b[1]]) = \
                (self.state[i][b[0],b[1]], self.state[i][a[0],a[1]])

    def contains(self, x):
        return self.down.down(x)

    def down_states(self, x, num=0):
        return [self.down.below.state[num][x] for x in self.contains(x)]

    def organs(self, x):
        return self.up.up(x)

    def up_states(self, x, num=0):
        return [self.up.above.state[num][x] for x in self.organs(x)]

    def next(self):
        exec self.on_next
        if not hasattr(self, "coords"):
            self.coords = [(x, y) for x in range(self.dim[0]) \
                               for y in range(self.dim[1])]
            random.shuffle(self.coords)
        if not self.inplace:
            ret = self.copy()
        else:
            ret = self
        for i in range(len(self.coords)):
            (x,y) = self.coords[i]
            rule = self.twoRule[self.rule[x, y]]
            if self.inplace:
                rule(self, (x,y))
            else:
                for (i, a) in enumerate(rule(self, (x,y))):
                    ret.state[i][x,y] = a
        return ret

    def states(self, x):
        nbh = self.twoNbh[self.nbh[x]]
        return ([self[x+n] for n in nbh], self.nbh[x], self.rule[x])

    def nbh_state(self, x, layer=0):
        nbh = self.twoNbh[self.nbh[x]]
        return statify([self.state[layer][(x+n)%self.width] for n in nbh], 
                       self.max_state)

    def sum(self, x, num=0):
        nbh = self.twoNbh[self.nbh[x]]
        layer = self.state[num]
        return sum([layer[(x+n)%self.width] for n in nbh])

    def outer_sum(self, x):
        return self.sum(x) - self[x]

    def random(self, nums=(0,), **kwargs):
        min = kwargs.get("min", 0)
        max = kwargs.get("max", self.max_state)
        partial = kwargs.get("partial", 100)
        if kwargs.get("nbh", True):
           self.nbh = np.random.randint(0, len(self.twoNbh), self.dim)
        if kwargs.get("rule", True):
            self.rule = np.random.randint(0, len(self.twoRule), self.dim)
        if partial == 100:
            for n in nums:
                self.state[n] = np.random.randint(min, max+1, self.dim)
            return
        elif partial == 0:
            for n in nums:
                self.state[n] = 0
            return
        else:
            for n in nums:
                indexes = np.random.randint(0, 101, self.dim) < partial
                choices = [self.state[n], 
                           np.random.randint(min, max+1, self.dim)]
                self.state[n] = np.choose(indexes, choices)
            

    def copy(self):
        ret = Lattice(self.twoNbh, self.twoRule, self.tmap, **self.kwargs)
        ret.coors = self.coords
        ret.up = self.up
        ret.down = self.down
        ret.rule = self.rule.copy()
        ret.nbh = self.nbh.copy()
        for i in range(len(self.state)):
            ret.state[i] = self.state[i].copy()
        return ret


class LatticeStack:
    def __init__(self, lattices, identifier=""):
        if not type(lattices) in (type(()), type([])):
            lattices = (lattices,)
        self.lattices = lattices
        self.identifier = identifier

    def __getitem__(self, t):
        return self.lattices[t]

    def next(self):
        new_lattices = []
        for l in self.lattices:
            new_lattices.append(l.next())
            if len(new_lattices) > 1:
                connect(new_lattices[-2], new_lattices[-1], l.down)
        return LatticeStack(new_lattices, self.identifier)

    def random(self, nums_arr=None, **kwargs):
        for i in range(len(self.lattices)):
            nums = (0,)
            if nums_arr != None and len(nums_arr) > i:
                nums = nums_arr[i]
            self.lattices[i].random(nums, **kwargs)

    def copy(self):
        ret = []
        for l in self.lattices:
            ret.append(l.copy())
            if len(ret) > 1:
                connect(ret[-2], ret[-1], l.down)
        return LatticeStack(ret, self.identifier)
               

class CA:
    def __init__(self, state, n=100, t=0):
        self.state = state

    def get(self):
        self.state = self.state.next()

    def set_state(self, state):
        self.state = state
        

class NullTMap:
    def __call__(self, lattice, x):
        return (0,0)

def tuple_add(a, b):
    return (a[0] + b[0], a[1] + b[1])

def tuple_div(a, b):
    return (a[0] / b, a[1] / b)

def tuple_mul(a, b):
    return (a[0] * b, a[1] * b)


def pick_dir(states, target_state):
    candidates = []
    i = 0
    while i < len(states):
        try:
            cand_i = states.index(target_state, i)
        except ValueError:
            break
        candidates.append(cand_i)
        i = cand_i + 1
    if len(candidates) == 0:
        return None
    return random.choice(candidates)

class WatorRule:
    def __init__(self, bf, bs, s, hs=None):
        self.bf = bf
        self.bs = bs
        self.s = s
        if hs == None:
            hs = s+1
        self.hs = hs

    def __call__(self, lattice, coord):
        HOCA = False
        if self.__class__.__name__ == "AdvWatorRule":
            HOCA = True

        self.lattice = lattice
        self.coord = coord
        self.dirs = lattice.twoNbh[lattice.nbh[coord]]
        lat = lattice.state[0]
        hunger = lattice.state[1]
        age = lattice.state[2]
        moved = lattice.state[3]
        if HOCA:
            blood = lattice.state[4]
        state = lat[coord]

        if state == 0:               # Just water
            return

        if moved[coord] == True:     # This fish has been moved already
            return

        age[coord] += 1 
        nbh = [lat[lattice.get_coord(tuple_add(coord, x))] for x in self.dirs]
        if state == 1:               # It's a fish
            dir = self.fish_swimming(nbh)
            if dir == None:
                return               # There's no place to swim to
            tar_coord = lattice.get_coord(tuple_add(coord, self.dirs[dir]))
            moved[coord] = True
            if age[coord] > self.bf: # Fish breeds
                age[tar_coord] = 0
                lat[tar_coord] = 1
                age[coord] = 0
            lattice.swap(tar_coord, coord)
            return

        # It's a shark
        if hunger[coord] >= self.s:   # It's dead
            age[coord] = 0
            lat[coord] = 0
            hunger[coord] = 0
            return
        hunger[coord] += 1           
        dir = pick_dir(nbh, 1)       # Look for food
        if dir == None:
            dir = self.shark_swimming(nbh)   # No food, just swim
            if dir == None:
                return               # Nowhere to swim to
        else:
            hunger[coord] = 0        # Hey, food! Eat.
            if HOCA:
                blood[coord] = True
            
        tar_coord = lattice.get_coord(tuple_add(coord, self.dirs[dir]))
        moved[coord] = True
        if age[coord] > self.bs and hunger[coord] < self.hs:     # Shark breeds
            lat[tar_coord] = 2
            age[tar_coord] = 0
            hunger[tar_coord] = 0
            # Random age after breeding
            age[coord] = 0
        else:                        # Shark just swims
            lat[tar_coord] = 0       # Maybe it eats a fish
        lattice.swap(tar_coord, coord) 
        return

    def fish_swimming(self, nbh):
        return pick_dir(nbh, 0)

    def shark_swimming(self, nbh):
        return pick_dir(nbh, 0)

class AdvWatorRule(WatorRule):
    def shark_swimming(self, nbh):
        ret = self.lattice.up_states(self.coord,1)[0]
        c = 0
        dir = 0
        for d in rot_pri:
            dir = d(self.dirs[ret])
            c = self.lattice.get_coord(tuple_add(self.coord,dir))
            if self.lattice.state[0][c] != 2:
                break
        if self.lattice.state[0][c] == 2:
            return None
        return self.dirs.index(dir)

def gradient_dir(states):
    max = 0
    min = 2700
    max_indices = []
    for (i, s) in enumerate(states):
        if s > max:
            max_indices = [i]
            max = s
        if s == max:
            max_indices.append(i)
        if s < min:
            min = s
    if min == max:
        return random.choice(range(1,5))
    return random.choice(max_indices)
    

class DiffuseRule:
    def __init__(self, decay=1.5):
        self.decay = decay

    def __call__(self, lattice, coord): 
        (x, y ) = coord
        (dx, dy) = [i*3 for i in coord]

        dirs = lattice.down.below.twoNbh[lattice.down.below.nbh[dx,dy]]
        state = lattice.state[0][x,y]
        nbh = [lattice.state[0][lattice.get_coord(tuple_add(coord,i))] \
                   for i in dirs]
        grad = gradient_dir(nbh)

        downsum = sum(lattice.down_states((x,y), 4))  # Sum of blood
        heresum = sum(nbh)                            # Sum of old blood
        newsum = heresum/(5*self.decay) + downsum*100
        return (newsum, grad)
        
    

class ModMap(null_map):
    def __init__(self, mod, **kwargs):
        super(ModMap, self).__init__(**kwargs)
        self.mod = mod

    def up(self, coord):
        return (tuple_div(coord, self.mod),)

    def down(self, coord):
        beg = tuple_mul(coord, self.mod)
        rnge = range(self.mod)
        return [self.below.get_coord(tuple_add(beg, (x,y))) \
                    for x in rnge for y in rnge]


def enlarge(data, factor):
    ret = np.zeros([factor*x for x in data.shape], data.dtype)
    for x in range(factor):
        for y in range(factor):
            ret[x::factor,y::factor] = data
    return ret

if __name__ == "__main__":
    states = [234,232,221,300]
    print gradient_dir(states)
    
