
#!/usr/bin/env python

#
# CA Explorer program for exploring cellular automata
# Copyright (C) 2009 Lauritz Thaulow
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

START_MESSAGE = """
  CA Explorer is Copyright (C) 2009 Lauritz Thaulow
  This program comes with ABSOLUTELY NO WARRANTY; for details,
  read the "About" dialog.

  This is free software, and you are welcome to redistribute it
  under certain conditions; see the "About" dialog for the license.
"""

VERSION = "0.1"

import wx
import os
import sys
from math import log10, ceil
from time import time, sleep
from types import TupleType, IntType, FloatType
import numpy as np
import WatorHOCA as hoca
from random import randint

ID_SAVE = 102
ID_ABOUT = 101
ID_EXIT = 110


class CA_wrapper:
    def __init__(self, state, ca_type="CA", write_mode="a"):
        self.state = state
        self.display = []
        r = state[0].twoRule[0]
        (w,h) = state[0].dim
        self.path = "data/%s_%dx%d_bf-%d_bs-%d_s-%d" % \
            (ca_type, w, h, r.bf, r.bs, r.s)
        self.log = open(self.path + ".dat", write_mode)
        titletempl = "WATOR-World with bf=%d, bs=%d and s=%d"
        titletempl += "\non a %dx%d lattice"
        d = {"caption": titletempl % (r.bf, r.bs, r.s, w, h),
             "path": self.path,
             "filename": os.path.basename(self.path)}
        template = open("plot_template.p", "r").read() % d
        open(self.path + ".p", "w").write(template)
        #print open("latex_template.tex", "r").read() % d
        rule = state[0].twoRule[0]
        state.random(rule=False, nbh=False, min=0, max=1, partial=20)
        state.random(rule=False, nbh=False, min=1, max=2, partial=15)

    def get_dim(self):
        return self.state[0].dim
    dim = property(get_dim)


class Display:
    def __init__(self, name, level, layer, eval_str):
        self.name = name
        self.level = level
        self.layer = layer
        self.eval = eval_str

    def get_data(self, wrapper):
        zoom = 1
        st = wrapper.state[self.level]
        while st.down.__class__.__name__ != "null_map":
            zoom *= st.down.mod
            st = st.down.below
        data = getattr(wrapper.state[self.level], self.name)
        if self.layer != None:
            data = data[self.layer]
        return (zoom, eval(self.eval))


def id_to_text(x):
    if type(x) == type(""):
        return x
    l = []
    keys = x.keys()
    keys.sort()
    for k in keys:
        l.append("%s: %s" % (k, x[k]))
    return ", ".join(l)

def id_to_filename(x):
    if type(x) == type(""):
        return x
    l = []
    keys = x.keys()
    keys.sort()
    for k in keys:
        l.append("%s-%s" % (k, x[k]))
    return "_".join(l)

class PatternPanel(wx.Panel):
    def __init__(self, parent, **kwargs):
        wx.Panel.__init__(self, parent, wx.ID_ANY, size=(440,680), style=wx.BG_STYLE_CUSTOM|wx.TAB_TRAVERSAL)
        self.parent = parent
        self.kwargs = kwargs
        self.SetBackgroundColour('white')
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.do_draw = True
        #self.Bind(wx.EVT_MOUSEWHEEL, self.OnWheel)
        #self.Bind(wx.EVT_MOTION, self.OnMouse)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.timer = wx.Timer()
        self.timer.SetOwner(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer)

        self.iterator = kwargs.get("iterator", iter([]))
        self.reset_cas()


    def reset_cas(self):
        self.t = 0
        self.cas = self.iterator.next()

        self.zoom = 2
        self.margin = 5
        self.sep = 5
        self.bot = 0
        self.end_t = self.kwargs.get("end_t", 1000)

        (self.w, self.h) = [self.zoom*x for x in self.cas[0].dim]
        self.w_sep = self.w + self.sep
        self.stripes = sum([len(x.display) for x in self.cas])
        self.w_tot = 2*self.margin + self.stripes*self.w + \
            (self.stripes-1)*self.sep
        self.h_tot = self.h + 2*self.margin + self.bot

        self.img = np.zeros((self.h_tot, self.w_tot, 3), dtype="uint8")
        self.img[...] = 255
        self.im = wx.EmptyImage(self.w_tot, self.h_tot)
        default_save = [1, 2, 20, 50, 100, 200, 400]
        self.save = self.kwargs.get("save", default_save)

        
                        
    def OnTimer(self, e):
        self.PaintPattern()

    def PaintPattern(self):
        if not self.do_draw:
            return
        self.t += 1

        dc = wx.BufferedDC(None, self.buf)
        col_no = 0
        for ca_wrapper in self.cas:
            lat = ca_wrapper.state[0].state[0]
            n_fish = np.sum(lat == 1)
            ca_wrapper.log.write("%d %d %d\n" % 
                                 (self.t, n_fish, np.sum(lat == 2)))
            max_fish = reduce(lambda x,y:x*y, ca_wrapper.state[0].dim)
            if n_fish in (0, max_fish):
                if self.end_t - self.t > 50:
                    self.end_t = self.t + 50
            ca_wrapper.state = ca_wrapper.state.next()
            for disp in ca_wrapper.display:
                (zoom, data) = disp.get_data(ca_wrapper)
                start = self.margin + col_no*self.w_sep
                stop = start + self.w
                for c in (0,1,2):
                    self.img[self.margin:self.margin+self.h,start:stop,c] = \
                        np.transpose(hoca.enlarge(data, self.zoom*zoom))
                col_no += 1
        self.im.SetData(self.img.tostring())
        bm = self.im.ConvertToBitmap()
        memdc = wx.MemoryDC(bm)
        if 0:
            memdc.SetBrush(wx.TRANSPARENT_BRUSH)
            col_no = 0
            for ca_wrapper in self.cas:
                curr_h = self.margin + self.h + 2
                num_in_box = len(ca_wrapper.display)
                for disp in ca_wrapper.display:
                    text = disp.name.capitalize()
                    (width, height) = memdc.GetTextExtent(text)
                    memdc.DrawText(text, self.margin + col_no*self.w_sep + \
                                       self.w/2 - width/2, curr_h)
                    col_no += 1
                text = id_to_text(ca_wrapper.state.identifier)
                (tw, th) = memdc.GetTextExtent(text)
                curr_h += th + 2
                left_edge = self.margin + (col_no-num_in_box)*self.w_sep 
                width = num_in_box*self.w_sep - self.sep
                middle = left_edge + self.width/2
                memdc.DrawText(text, middle - tw/2, curr_h)
                memdc.DrawRectangle(left_edge-4,self.margin-4,
                                    width+8,self.h_tot-self.margin-\
                                        2*self.margin+8)

        dc.Blit(0, 0, self.w_tot, self.h_tot, memdc, 0, 0)
        del dc
        self.OnPaint()
        identifier = id_to_filename(self.cas[-1].state.identifier)
        if self.t in self.save:
            tosave = bm.ConvertToImage()
            tosave.SetOption("quality", "85")
            tosave.SaveFile(self.cas[0].path + "_t-%d.png" % (self.t,),
                            wx.BITMAP_TYPE_PNG)
        if self.t in (self.end_t, 1000):
            [x.log.write("\n\n") for x in self.cas]
            self.end_t = self.kwargs.get("end_t", 1000)
            self.t = 0
            try:
                self.reset_cas()
            except StopIteration:
                self.do_draw = False
                self.parent.OnExit(None)
                return
        self.timer.Start(10)

    def updateSize(self):
        (self.width, self.height) = self.GetClientSizeTuple()
        return (self.width, self.height)

    def OnResize(self, evt):
        width, height = self.updateSize()
        self.buf = wx.EmptyBitmap(width, height)
        dc = wx.BufferedDC(None, self.buf)
        dc.Clear()
        del dc
        self.PaintPattern()
        #self.Reset()

    def OnPaint(self, evt=None):
        if not hasattr(self, "buf"):
            self.OnResize(None)
        wx.BufferedPaintDC(self, self.buf)

    def OnChar(self, e):
        kc = e.GetKeyCode()
        print kc
        if kc == 32:
            self.h = 500
        if kc in (65, 97):  # Aa
            (self.initial, self.display) = \
                my_example1()
        if kc in (66, 98):  # Bb
            (self.initial, self.display) = \
                my_example2()
        self.PaintPattern()
        

class AboutDialog(wx.Dialog):
    def __init__(self):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "About CA Explorer", 
                           size=(640, 480), style=wx.DEFAULT_DIALOG_STYLE)
        sizer = wx.BoxSizer(wx.VERTICAL)
        text = "\nCA Explorer %s is (C) 2009 Lauritz Thaulow\n" % VERSION
        sizer.Add(wx.StaticText(self, wx.ID_ANY, text))
        path = os.path.dirname(sys.argv[0])
        license = open(os.path.join(path, "LICENSE.txt"), "r").read()
        content = wx.TextCtrl(self, wx.ID_ANY, license, (0,0), (640, -1),
                              style=wx.TE_READONLY|wx.TE_MULTILINE)
        content.SetFont(wx.Font(10, wx.FONTFAMILY_TELETYPE, 
                                wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL))
        sizer.Add(content, 1, wx.EXPAND)
        ok = wx.Button(self, wx.ID_ANY, "OK")
        self.Bind(wx.EVT_BUTTON, self.OnOK)
        sizer.Add(ok, 0, wx.ALIGN_CENTER)
        self.SetSizer(sizer)
        content.SetInsertionPoint(0)
        
    def OnOK(self, evt):
        self.Destroy()


class RectDialog(wx.Dialog):
    def __init__(self, vars):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "Specify dimensions", 
                           size=(640, 480), 
                           style=wx.DEFAULT_DIALOG_STYLE)
        self.vars = vars
        self.inp = {}
        sizer = wx.BoxSizer(wx.VERTICAL)
        for v in vars:
            sizer2 = wx.BoxSizer(wx.HORIZONTAL)            
            sizer2.Add(wx.StaticText(self, wx.ID_ANY, "%s: " % v))
            self.inp[v] = wx.TextCtrl(self, wx.ID_ANY, "%d" % self.vars[v])
            sizer2.Add(self.inp[v])
            sizer.Add(sizer2)
        ok = wx.Button(self, wx.ID_ANY, "OK")
        self.Bind(wx.EVT_BUTTON, self.OnOK)
        sizer.Add(ok, 0, wx.ALIGN_CENTER)
        self.SetSizer(sizer)
        
    def OnOK(self, evt):
        for v in self.vars:
            self.vars[v] = int(self.inp[v].GetValue())
        self.Destroy()


class MainWindow(wx.Frame):
    def __init__(self, parent, id, title, **kwargs):
        wx.Frame.__init__(self, parent, id, title, size=(1000, 680))
        self.args = {}
        self.changed = False
        self.curr_clss = None
        filemenu= wx.Menu()
        filemenu.Append(ID_ABOUT, "&About"," Information about this program")
        filemenu.AppendSeparator()
        filemenu.Append(ID_EXIT,"&Quit"," Exit the program.")
        menuBar = wx.MenuBar()
        menuBar.Append(filemenu,"&File")
        self.SetMenuBar(menuBar)

        self.pp = PatternPanel(self, **kwargs)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.pp, 1, wx.EXPAND)
        self.SetAutoLayout(True)
        self.SetSizer(self.sizer)
        self.Layout()
        #self.sizer.Fit(self)

        self.SetMinSize(wx.Size(10,10))
        self.CreateStatusBar()
                  
        wx.EVT_MENU(self, ID_ABOUT, self.OnAbout)
        wx.EVT_MENU(self, ID_EXIT, self.OnExit)
        #self.Bind(wx.EVT_IDLE, self.OnIdle)

        self.Show(True)
        self.pp.Bind(wx.EVT_SIZE, self.pp.OnResize)
        self.SetFocus()
        #self.pp.OnResize(None)
        #self.pp.PaintPattern(self.args)
        wx.EVT_CLOSE(self, self.OnClose)

    def OnAbout(self, e):
        d = AboutDialog()
        d.ShowModal()
        d.Destroy()

    def OnExit(self, e):
        self.Close(True)  # Close the frame.

    def OnClose(self, e):
        gnuplot = self.pp.kwargs.get("gnuplot", "")
        if gnuplot:
            import subprocess as sp
            masterdir = "./"
            for caw in self.pp.cas:
                caw.log.close()
                r = sp.call([gnuplot,
                             "%s.p" % (os.path.basename(caw.path),)], 
                            cwd="%sdata" % (masterdir,))
        self.pp.do_draw = False
        self.Destroy()



def start(**kwargs):
    print START_MESSAGE
    import sys,os
    if not kwargs.has_key("gnuplot"):
        kwargs["gnuplot"] = ""
    app = wx.PySimpleApp()
    frame = MainWindow(None, wx.ID_ANY, "WATOR-World", **kwargs)
    app.MainLoop()


if __name__ == '__main__':
    #import cProfile
    #cProfile.run('start()', 'profile.txt')
    start()
    


